# -*- coding: UTF-8 -*-
from .braille import *
from .ipsum import *
from .leetspeak import *
from .morse import *
from .navajo import *
from .radio import *
from .southpark import *
from .tomtom import *




# -*- coding: UTF-8 -*-
"""LZ78 Codec - Lempel-Ziv 1978 compression algorithm.

NB: Not an encoding properly speaking.

This codec:
- en/decodes strings from str to str
- en/decodes strings from bytes to bytes
- decodes file content to str (read)
- encodes file content from str to bytes (write)

Inspired from: https://github.com/mileswatson/lempel-ziv-compression
"""
from ..__common__ import *


__examples__ = {'enc-dec(lz78)': ["test", "This is a test", "@random{512,1024,2048}"]}


def lz78_compress(input, errors="strict"):
    """ Compresses the given data by applying LZ78 compression algorithm. """
    data = tuple(c if isinstance(c, int) else ord(c) for c in input)
    if len(data) == 0:
        return "", 0
    out = (data[0], )
    d = {tuple(): (0, ), (data[0], ): (1, )}
    a, b, ctr = 1, 1, [2]
    while b < len(data):
        if not data[a:b+1] in d:
            w = d[data[a:b]]
            out += w + tuple(0 for i in range(len(ctr) - len(w) - int(sum(ctr) == 1))) + (data[b], )
            d[data[a:b+1]] = tuple(ctr)
            for i in range(len(ctr)):
                ctr[i] += 1
                if ctr[i] != 256:
                    break
                else:
                    ctr[i] = 0
                    if i == len(ctr) - 1:
                        ctr.append(1)
            a = b + 1
        b += 1
    if data[a:b] in d and a != b:
        w = tuple(d[data[a:b]])
        out += w + tuple(0 for i in range(len(ctr) - len(w)))
    return "".join(chr(i) for i in out), len(out)


def lz78_decompress(input, errors="strict"):
    """ Decompresses the given data. """
    data = tuple(c if isinstance(c, int) else ord(c) for c in input)
    if len(data) == 0:
        return "", 0
    out = (data[0], )
    l = [tuple(), out]
    a, b, c, i, char = 1, 1, 256, 0, False
    try:
        while a < len(data):
            if char:
                out += (data[a], )
                l.append(l[i] + (data[a], ))
                char = False
                a += 1
                if len(l) == c + 1:
                    b += 1
                    c *= 256
            else:
                i, m = 0, 1
                for j in range(b):
                    i += data[a + j] * m
                    m *= 256
                out += l[i]
                a += b
                char = True
    except:
        return handle_error("lz78", errors, decode=True)(chr(data[a]), a), len(input)
    return "".join(chr(i) for i in out), len(out)


add("lz78", lz78_compress, lz78_decompress)




# -*- coding: UTF-8 -*-
"""Base45 Codec - base45 content encoding.

This codec:
- en/decodes strings from str to str
- en/decodes strings from bytes to bytes
- decodes file content to str (read)
- encodes file content from str to bytes (write)
"""
from ._base import digits, lower, main, upper
from ..__common__ import *


__examples__ = {
    'enc(base45|base-45|base_45)': {'this is a test!': "AWE+EDH44.OEOCC7WE QEX0"},
    'dec(base45)':                 {'BAD STRING\00': None, 'AWE+EDH44.OEOCC7WE QEX000': None},
}
__guess__ = ["base45", "base45-inv"]


B45 = {
    '':    digits + upper + " $%*+-./:",
    'inv': upper + digits + " $%*+-./:",
}


__chr = lambda c: chr(c) if isinstance(c, int) else c
__ord = lambda c: ord(c) if not isinstance(c, int) else c


class Base45DecodeError(ValueError):
    pass


def base45_encode(mode):
    mode = mode.replace("inverted", "inv").replace("_", "-").lstrip("-")
    b45 = B45[['inv', ''][mode == ""]]
    def encode(text, errors="strict"):
        t, s = b(text), ""
        for i in range(0, len(text), 2):
            n = 256 * __ord(t[i])
            try:
                n += __ord(t[i+1])
            except IndexError:
                n = __ord(t[i])
                s += b45[n % 45] + b45[n // 45]
                break
            m = n // 45**2
            n -= m * 45**2
            s += b45[n % 45] + b45[n // 45] + b45[m]
        return s, len(text)
    return encode


def base45_decode(mode):
    mode = mode.replace("inverted", "inv").replace("_", "-").lstrip("-")
    b45 = {c: i for i, c in enumerate(B45[['inv', ''][mode == ""]])}
    def decode(text, errors="strict"):
        t, s, err = b(text), "", "'base45' codec can't decode character '%s' in position %d"
        ehandler = handle_error("base45", errors, Base45DecodeError, decode=True)
        for i in range(0, len(text), 3):
            try:
                n = b45[__chr(t[i])]
            except KeyError:
                ehandler(__chr(t[i]), i, s)
            try:
                j = i + 1
                n += 45 * b45[__chr(t[j])]
            except KeyError:
                ehandler(__chr(t[j]), j, s)
            except IndexError:
                ehandler(__chr(t[i]), i, s)
            try:
                k = i + 2
                n += 45 ** 2 * b45[__chr(t[k])]
            except KeyError:
                ehandler(__chr(t[k]), k, s)
            except IndexError:
                s += __chr(n)
                continue
            s += __chr(n // 256) + __chr(n % 256)
        return s, len(text)
    return decode


add("base45", base45_encode, base45_decode, r"^base[-_]?45(|[-_]inv(?:erted)?)$")
main = main(45, "<https://datatracker.ietf.org/doc/html/draft-faltstrom-base45-04.txt>")




# -*- coding: UTF-8 -*-
"""BaseN functions with N a power of 2.

"""
from math import ceil, log

from ..__common__ import *
from ._base import base, _get_charset, BaseError


# base en/decoding functions for N a power of 2
class Base2NError(BaseError):
    pass


class Base2NDecodeError(BaseError):
    pass


class Base2NEncodeError(BaseError):
    pass


def base2n(charset, pattern=None, name=None, **kwargs):
    """ Base-N codec factory for N a power of 2.
    
    :param charset: charset selection function
    :param pattern: matching pattern for the codec name (first capturing group is used as the parameter for selecting
                     the charset)
    :param name:    forced encoding name (useful e.g. for zbase32)
    """
    base(charset, pattern, True, base2n_encode, base2n_decode, name, **kwargs)


def base2n_encode(string, charset, errors="strict", exc=Base2NEncodeError):
    """ 8-bits characters to base-N encoding for N a power of 2.
    
    :param string:  string to be decoded
    :param charset: base-N characters set
    :param errors:  errors handling marker
    :param exc:     exception to be raised in case of error
    """
    bs, r, n = "", "", len(charset)
    # find the number of bits for the given character set and the quantum
    nb_out = int(log(n, 2))
    q = nb_out
    while q % 8 != 0:
        q += nb_out
    # iterate over the characters, gathering bits to be mapped to the charset
    for i, c in enumerate(string):
        c = c if isinstance(c, int) else ord(c)
        bs += "{:0>8}".format(bin(c)[2:])
        while len(bs) >= nb_out:
            r += charset[int(bs[:nb_out], 2)]
            bs = bs[nb_out:]
    if len(bs) > 0:
        for i in range(0, len(bs), nb_out):
            c = ("{:0<%d}" % nb_out).format(bs[i:i+nb_out])
            p = len(c) - len(bs[i:i+nb_out])
            r += charset[int(c, 2)]
    l = len(r) * nb_out
    while l % q != 0:
        l += nb_out
    return r + int(l / nb_out - len(r)) * "="


def base2n_decode(string, charset, errors="strict", exc=Base2NDecodeError):
    """ Base-N to 8-bits characters decoding for N a power of 2.
    
    :param string:  string to be decoded
    :param charset: base-N characters set
    :param errors:  errors handling marker
    :param exc:     exception to be raised in case of error
    """
    bs, r, n = "", "", len(charset)
    # particular case: for hex, ensure the right case in the charset ; not that this way, if mixed cases are used, it
    #                   will trigger an error (this is the expected behavior)
    if n == 16:
        if any(c in string for c in "abcdef"):
            charset = charset.lower()
        elif any(c in string for c in "ABCDEF"):
            charset = charset.upper()
    string = re.sub(r"\s", "", string)
    # find the number of bits for the given character set and the number of padding characters
    nb_in = int(log(n, 2))
    n_pad = len(string) - len(string.rstrip("="))
    # iterate over the characters, mapping them to the character set and converting the resulting bits to 8-bits chars
    for i, c in enumerate(string):
        if c == "=":
            bs += "0" * nb_in
        else:
            try:
                bs += ("{:0>%d}" % nb_in).format(bin(charset.index(c))[2:])
            except ValueError:
                if errors == "strict":
                    raise exc("'base' codec can't decode character '{}' in position {}".format(c, i))
                elif errors == "replace":
                    bs += "0" * nb_in
                elif errors == "ignore":
                    continue
                else:
                    raise ValueError("Unsupported error handling {}".format(errors))
        if len(bs) > 8:
            r += chr(int(bs[:8], 2))
            bs = bs[8:]
    # if the number of bits is not multiple of 8 bits, it could mean a bad padding
    if len(bs) != 8:
        if errors == "strict":
            raise Base2NDecodeError("Incorrect padding")
        elif errors in ["replace", "ignore"]:
            pass
        else:
            raise ValueError("Unsupported error handling {}".format(errors))
    r += chr(int(bs, 2))
    np = int(ceil(n_pad * nb_in / 8.0))
    return r[:-np] if np > 0 else r




# -*- coding: UTF-8 -*-
"""Generic baseN functions.

"""
from argparse import ArgumentParser, RawTextHelpFormatter
from math import log
from six import integer_types, string_types
from string import ascii_lowercase as lower, ascii_uppercase as upper, digits, printable
from textwrap import wrap
from types import FunctionType, MethodType

from ..__common__ import *
from ..__info__ import __version__


class BaseError(ValueError):
    pass


class BaseDecodeError(BaseError):
    pass


class BaseEncodeError(BaseError):
    pass


def _generate_charset(n):
    """ Generate a characters set.
    
    :param n: size of charset
    """
    if 1 < n <= len(printable):
        return printable[:n]
    elif len(printable) < n < 256:
        return "".join(chr(i) for i in range(n))
    raise ValueError("Bad size of character set")


def _get_charset(charset, p=""):
    """ Characters set selection function. It allows to define charsets in many different ways.
    
    :param charset: charset object, can be a string (the charset itself), a function (that chooses the right charset
                     depending on the input parameter) or a dictionary (either by exact key or by pattern matching)
    :param p:       the parameter for choosing the charset
    """
    # case 1: charset is a function, so return its result
    if isinstance(charset, FunctionType):
        return charset(p)
    # case 2: charset is a string, so return it
    elif isinstance(charset, string_types):
        return charset
    # case 3: charset is a dict with keys '' and 'inv', typically for a charset using lowercase and uppercase characters
    #          that can be inverted
    elif isinstance(charset, dict) and list(charset.keys()) == ["", "inv"]:
        return charset["inv" if re.match(r"[-_]inv(erted)?$", p) else ""]
    # case 4: charset is a dict, but not with the specific keys '' and 'inv', so consider it as pattern-charset pairs
    elif isinstance(charset, dict):
        # try to handle [p]arameter as a simple key
        try:
            return charset[p]
        except KeyError:
            pass
        # or handle [p]arameter as a pattern
        default, n = None, None
        for pattern, cset in charset.items():
            n = len(cset)
            if pattern == "":
                default = cset
                continue
            if re.match(pattern, p):
                return cset
        # special case: the given [p]arameter can be the charset itself if it has the right length
        p = re.sub(r"^[-_]+", "", p)
        if len(p) == n:
            return p
        # or simply rely on key ''
        if default is not None:
            return default
    raise ValueError("Bad charset descriptor")


# generic base en/decoding functions
def base_encode(input, charset, errors="strict", exc=BaseEncodeError):
    """ Base-10 to base-N encoding.
    
    :param input:   input (str or int) to be decoded
    :param charset: base-N characters set
    :param errors:  errors handling marker
    :param exc:     exception to be raised in case of error
    """
    i, n, r = input if isinstance(input, integer_types) else s2i(input), len(charset), ""
    while i > 0:
        i, c = divmod(i, n)
        r = charset[c] + r
    return r


def base_decode(input, charset, errors="strict", exc=BaseDecodeError):
    """ Base-N to base-10 decoding.
    
    :param input:   input to be decoded
    :param charset: base-N characters set
    :param errors:  errors handling marker
    :param exc:     exception to be raised in case of error
    """
    i, n, dec = 0, len(charset), lambda n: base_encode(n, [chr(x) for x in range(256)], errors, exc)
    for k, c in enumerate(input):
        try:
            i = i * n + charset.index(c)
        except ValueError:
            handle_error("base", errors, exc, decode=True)(c, k, dec(i))
    return dec(i)


# base codec factory functions
def base(charset, pattern, pow2=False, encode_template=base_encode, decode_template=base_decode, name=None, **kwargs):
    """ Base-N codec factory.
    
    :param charset: charset selection function
    :param pattern: matching pattern for the codec name (first capturing group is used as the parameter for selecting
                     the charset)
    :param pow2:    whether the base codec's N is a power of 2
    """
    n = len(_get_charset(charset))
    nb = log(n, 2)
    if pow2 and nb != int(nb):
        raise BaseError("Bad charset ; {} is not a power of 2".format(n))
    
    def encode(param=""):
        a = _get_charset(charset, param)
        def _encode(input, errors="strict"):
            return encode_template(input, a, errors), len(input)
        return _encode
    
    def decode(param=""):
        a = _get_charset(charset, param)
        sl, sc = "\n" not in a, "\n" not in a and not "\r" in a
        def _decode(input, errors="strict"):
            input = _stripl(input, sc, sl)
            return decode_template(input, a, errors), len(input)
        return _decode
    
    kwargs['len_charset'] = n
    kwargs['printables_rate'] = 1.
    n = "base{}".format(n) if name is None else name
    kwargs['guess'] = kwargs.get('guess', [n])
    add(n, encode, decode, pattern, entropy=nb, **kwargs)


def base_generic():
    """ Base-N generic codec. """
    def encode(n):
        a = _generate_charset(int(n))
        def _encode(input, errors="strict"):
            return base_encode(input, a, errors), len(input)
        return _encode
    
    def decode(n):
        a = _generate_charset(int(n))
        sl, sc = "\n" not in a, "\n" not in a and not "\r" in a
        def _decode(input, errors="strict"):
            input = _stripl(input, sc, sl)
            return base_decode(input, a, errors), len(input)
        return _decode
    
    add("base", encode, decode, r"^base[-_]?([2-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])(?:[-_]generic)?$",
        guess=["base%d-generic" % i for i in range(2, 255)], entropy=lambda e, n: log(int(n.split("-")[0][4:]), 2),
        len_charset=lambda n: int(n.split("-")[0][4:]), printables_rate=1., category="base-generic", penalty=.4)


def main(n, ref=None, alt=None, inv=True):
    base = str(n) + ("-" + alt.lstrip("-") if alt else "")
    src = "The data are encoded as described for the base%(base)s alphabet in %(reference)s.\n" % \
          {'base': base, 'reference': "\n" + ref if len(ref) > 20 else ref} if ref else ""
    descr = """Usage: base%(base)s [OPTION]... [FILE]
Base%(base)s encode or decode FILE, or standard input, to standard output.

With no FILE, or when FILE is -, read standard input.

Mandatory arguments to long options are mandatory for short options too.
  -d, --decode          decode data
  -i, --ignore-garbage  when decoding, ignore non-alphabet characters
%(inv)s  -w, --wrap=COLS       wrap encoded lines after COLS character (default 76).
                          Use 0 to disable line wrapping

      --help     display this help and exit
      --version  output version information and exit

%(source)sWhen decoding, the input may contain newlines in addition to the bytes of
the formal base%(base)s alphabet.  Use --ignore-garbage to attempt to recover
from any other non-alphabet bytes in the encoded stream.

Report base%(base)s translation bugs to <https://github.com/dhondta/python-codext/issues/new>
Full documentation at: <https://python-codext.readthedocs.io/en/latest/enc/base.html>
""" % {'base': base, 'source': src,
       'inv': ["", "  -I, --invert          invert charsets from the base alphabet (e.g. lower- and uppercase)\n"][inv]}
    
    def _main():
        p = ArgumentParser(description=descr, formatter_class=RawTextHelpFormatter, add_help=False)
        p.format_help = MethodType(lambda s: s.description, p)
        p.add_argument("file", nargs="?")
        p.add_argument("-d", "--decode", action="store_true")
        p.add_argument("-i", "--ignore-garbage", action="store_true")
        if inv:
            p.add_argument("-I", "--invert", action="store_true")
        p.add_argument("-w", "--wrap", type=int, default=76)
        p.add_argument("--help", action="help")
        p.add_argument("--version", action="version")
        p.version = "CodExt " + __version__
        args = p.parse_args()
        args.invert = getattr(args, "invert", False)
        c, f = _input(args.file), [encode, decode][args.decode]
        c = c.rstrip("\r\n") if isinstance(c, str) else c.rstrip(b"\r\n")
        try:
            c = f(c, "base" + base + ["", "-inv"][args.invert], ["strict", "ignore"][args.ignore_garbage])
        except Exception as err:
            print("%sbase%s: invalid input" % (getattr(err, "output", ""), base))
            return 1
        for l in wrap(ensure_str(c), args.wrap):
            print(l)
        return 0
    return _main




# -*- coding: UTF-8 -*-
"""Pkzip Codec - pkzip content compression.

NB: Not an encoding properly speaking.

This codec:
- en/decodes strings from str to str
- en/decodes strings from bytes to bytes
- decodes file content to str (read)
- encodes file content from str to bytes (write)
"""
import zipfile

from ..__common__ import *


_str          = ["test", "This is a test", "@random{512,1024,2048}"]
__examples1__ = {'enc-dec(pkzip-deflate|deflate)': _str}
__examples2__ = {'enc-dec(pkzip_bz2|bzip2)':       _str}
__examples3__ = {'enc-dec(pkzip-lzma|lzma)':       _str}


if PY3:
    def pkzip_encode(compression_type):
        def _encode(text, errors="strict"):
            c = zipfile._get_compressor(compression_type)
            return c.compress(b(text)) + c.flush(), len(text)
        return _encode


    def pkzip_decode(compression_type):
        def _decode(data, errors="strict"):
            d = zipfile._get_decompressor(compression_type)
            r = d.decompress(b(data))
            return r, len(r)
        return _decode


    add("pkzip_deflate", pkzip_encode(8), pkzip_decode(8), r"(?:(?:pk)?zip[-_])?deflate",
        entropy=7.9, examples=__examples1__, guess=["deflate"])

    add("pkzip_bzip2", pkzip_encode(12), pkzip_decode(12), r"(?:(?:pk)?zip[-_])?bz(?:ip)?2",
        entropy=7.9, examples=__examples2__, guess=["bz2"])

    add("pkzip_lzma", pkzip_encode(14), pkzip_decode(14), r"(?:(?:pk)?zip[-_])?lzma",
        entropy=7.9, examples=__examples3__, guess=["lzma"])




# -*- coding: UTF-8 -*-
"""Manchester Codec - Manchester content encoding.

This codec:
- en/decodes strings from str to str
- en/decodes strings from bytes to bytes
- decodes file content to str (read)
- encodes file content from str to bytes (write)
"""
from ..__common__ import *


__examples1__ = {'enc(manchester)': {'This is a test!': "fei\x95i\x96jZYUi\x96jZYUiVYUjeifjZjeYV"}}
__examples2__ = {
    'enc(manchester-inverted|ethernet|ieee802.4)': {
        'This is a test!': "\x99\x9a\x96j\x96i\x95\xa5\xa6\xaa\x96i\x95\xa5\xa6\xaa\x96\xa9\xa6\xaa\x95\x9a\x96\x99"
                           "\x95\xa5\x95\x9a\xa6\xa9",
    },
}


def manchester_encode(clock):
    def encode(text, errors="strict"):
        r = ""
        for c in text:
            bin_c = bin(ord(c))[2:].zfill(8)
            for i in range(0, 8, 4):
                r += chr(int("".join(2*bit for bit in bin_c[i:i+4]), 2) ^ clock)
        return r, len(b(text))
    return encode


def manchester_decode(clock):
    def decode(text, errors="strict"):
        r, bits = "", ""
        for c in text:
            bin_c = bin(ord(c) ^ clock)[2:].zfill(8)
            bits += "".join(bin_c[i] for i in range(0, len(bin_c), 2))
            if len(bits) == 8:
                r += chr(int(bits, 2))
                bits = ""
        return r, len(b(text))
    return decode


add("manchester", manchester_encode(0x55), manchester_decode(0x55), examples=__examples1__, printables_rate=.25,
    entropy=lambda e: .17616*e+2.56229)
add("manchester-inverted", manchester_encode(0xaa), manchester_decode(0xaa), examples=__examples2__,
    pattern=r"^(?:manchester-inverted|ethernet|ieee802\.4)$", entropy=lambda e: .17616*e+2.56229)




# -*- coding: UTF-8 -*-
from .affine import *
from .atbash import *
from .bacon import *
from .barbie import *
from .citrix import *
from .rot import *
from .scytale import *
from .shift import *
from .xor import *




# -*- coding: UTF-8 -*-
"""Letters Codec - letter indices-related content encoding.

This codec:
- en/decodes strings from str to str
- en/decodes strings from bytes to bytes
- decodes file content to str (read)
- encodes file content from str to bytes (write)
"""
import random

from ..__common__ import *


__examples__ = {
    'enc-dec(ipsum|lorem-ipsum)': ["This is a test."],
    'enc(ipsum)':                 {'Bad test#': None},
}


DICT = {
    'a': ['a', 'ac', 'accumsan', 'ad', 'adipiscing', 'aenean', 'aliquam', 'aliquet', 'amet', 'ante', 'aptent', 'arcu',
          'at', 'auctor', 'augue'],
    'b': ['babel', 'bibendum', 'blandit', 'bomba', 'botum', 'buxus'],
    'c': ['class', 'commodo', 'condimentum', 'congue', 'consectetur', 'consequat', 'conubia', 'convallis', 'cras',
          'cubilia', 'curabitur', 'curae', 'cursus'],
    'd': ['dapibus', 'diam', 'dictum', 'dictumst', 'dignissim', 'dis', 'dolor', 'donec', 'dui', 'duis'],
    'e': ['efficitur', 'egestas', 'eget', 'eleifend', 'elementum', 'elit', 'enim', 'erat', 'eros', 'est', 'et', 'etiam',
          'eu', 'euismod', 'ex'],
    'f': ['facilisis', 'fames', 'faucibus', 'felis', 'fermentum', 'feugiat', 'finibus', 'fringilla', 'fusce'],
    'g': ['gadus', 'galliarus', 'ganeo', 'garba', 'gemma', 'gener', 'genuine', 'gestus', 'gramma', 'gravida', 'grex',
          'gusto', 'guttur', 'gyro'],
    'h': ['habitant', 'habitasse', 'hac', 'haicu', 'halo', 'helleborum', 'hendrerit', 'hilarius', 'himenaeos',
          'horreum', 'hydrus', 'hystericus'],
    'i': ['iaculis', 'id', 'imperdiet', 'in', 'inceptos', 'integer', 'interdum', 'ipsum'],
    'j': ['jaccae', 'jacio', 'jecur', 'jocundiatas', 'jovis', 'juctim', 'juger', 'juno', 'jussum', 'justo'],
    'k': ['kal', 'kalatorium', 'kalium', 'kaput', 'kardo', 'kenia', 'koppa', 'kum'],
    'l': ['lacinia', 'lacus', 'laoreet', 'lectus', 'leo', 'libero', 'ligula', 'litora', 'lobortis', 'lorem', 'luctus'],
    'm': ['maecenas', 'magna', 'magnis', 'malesuada', 'massa', 'mattis', 'mauris', 'maximus', 'metus', 'mi', 'molestie',
          'mollis', 'montes', 'morbi', 'mus'],
    'n': ['nam', 'nascetur', 'natoque', 'nec', 'neque', 'netus', 'nibh', 'nisi', 'nisl', 'non', 'nostra', 'nulla',
          'nullam', 'nunc'],
    'o': ['odio', 'orci', 'ornare'],
    'p': ['parturient', 'pellentesque', 'penatibus', 'per', 'pharetra', 'phasellus', 'placerat', 'platea', 'porta',
          'porttitor', 'posuere', 'potenti', 'praesent', 'pretium', 'primis', 'proin', 'pulvinar', 'purus'],
    'q': ['qua', 'quadrum', 'quam', 'quasi', 'quintum', 'quis', 'quisque', 'quo', 'quom', 'quota', 'qur'],
    'r': ['radicitus', 'radius', 'ratio', 'recidivus', 'rectio', 'rhoncus', 'ridiculus', 'risus', 'ros', 'rutrum'],
    's': ['sagittis', 'sapien', 'scelerisque', 'sed', 'sem', 'semper', 'senectus', 'sit', 'sociosqu', 'sodales',
          'sollicitudin', 'suscipit', 'suspendisse'],
    't': ['taciti', 'tellus', 'tempor', 'tempus', 'tincidunt', 'torquent', 'tortor', 'tristique', 'turpis'],
    'u': ['ullamcorper', 'ultrices', 'ultricies', 'urna', 'ut'],
    'v': ['varius', 'vehicula', 'vel', 'velit', 'venenatis', 'vestibulum', 'vitae', 'vivamus', 'volutpat', 'vulputate'],
    'w': ['wadiarus', 'warantus', 'warra', 'werumensium', 'wormicia'],
    'x': ['xandicus', 'xenon', 'xenium', 'xiphias', 'xvir', 'xylon', 'xysticus', 'xystus'],
    'y': ['yata', 'yatum', 'yatus', 'ypra'],
    'z': ['zamia', 'zelosus', 'zerum', 'zonatus', 'zymus'],
}
SCHARS = ".,:;+=-*/\\ "


def ipsum_encode(text, errors="strict"):
    s = ""
    for i, c in enumerate(text):
        try:
            if c in SCHARS:
                s = s.rstrip(" ") + c + " "
            else:
                w = random.choice(DICT[c.lower()])
                s += (w.capitalize() if c.isupper() else w) + " "
        except KeyError:
            s += handle_error("ipsum", errors, " ")(c, i)
    return s[:-1], len(text)


def ipsum_decode(text, errors="strict"):
    s = ""
    words = text.split(" ")
    for i, w in enumerate(words[:-1] if words[-1] == "" else words):
        if w.strip() == "":
            s += " "
        elif w in SCHARS:
            s += w
        else:
            try:
                DICT[w[0].lower()]
                s += w[0]
            except KeyError:
                s += handle_error("ipsum", errors, decode=True, item="word")(w, i)
            if w[-1] in SCHARS:
                s += w[-1]
    return s, len(text)


add("ipsum", ipsum_encode, ipsum_decode, pattern=r"^(?:lorem[-_]?)?ipsum$", printables_rate=1.)




# -*- coding: UTF-8 -*-
"""BCD Codec - Binary Coded Decimal content encoding.

This codec:
- en/decodes strings from str to str
- en/decodes strings from bytes to bytes
- decodes file content to str (read)
- encodes file content from str to bytes (write)
"""
from ..__common__ import *


__examples1__ = {
    'enc(bcd|binary-coded-decimal|binary_coded_decimal)': {
        'This is a test!': "\x08A\x04\x10Q\x15\x03!\x05\x11P2\tp2\x11a\x01\x11Q\x16\x030",
   },
    'dec(binary-coded-decimal)': {
        '\xaf':                                                         None,
        '\xff':                                                         None,
        '\x08A\x04\x10Q\x15\x03!\x05\x11P2\tp2\x11a\x01\x11Q\x16\x030': "This is a test!",
   },
}
__examples2__ = {
    'enc(bcd-ext0|bcd_extended_zeros)': {
        'This is a test': "\x00\x08\x04\x01\x00\x04\x01\x00\x05\x01\x01\x05\x00\x03\x02\x01\x00\x05\x01\x01\x05\x00"
                          "\x03\x02\x00\t\x07\x00\x03\x02\x01\x01\x06\x01\x00\x01\x01\x01\x05\x01\x01\x06\x00",
   },
}
__examples3__ = {
    'enc(bcd-ext1|bcd_extended_ones)': {
        'This is a test': "\xf0\xf8\xf4\xf1\xf0\xf4\xf1\xf0\xf5\xf1\xf1\xf5\xf0\xf3\xf2\xf1\xf0\xf5\xf1\xf1\xf5\xf0"
                          "\xf3\xf2\xf0\xf9\xf7\xf0\xf3\xf2\xf1\xf1\xf6\xf1\xf0\xf1\xf1\xf1\xf5\xf1\xf1\xf6\xf0",
   },
}


CODE = {str(i): bin(i)[2:].zfill(4) for i in range(10)}


def bcd_encode(prefix=""):
    def encode(text, errors="strict"):
        r, bits = "", prefix
        for c in text:
            for i in str(ord(c)).zfill(3):
                bits += CODE[i]
                if len(bits) == 8:
                    r += chr(int(bits, 2))
                    bits = prefix
        if len(bits) > 0:
            r += chr(int(bits + "0000", 2))
        return r, len(b(text))
    return encode


def bcd_decode(prefix=""):
    def decode(text, errors="strict"):
        code = {v: k for k, v in CODE.items()}
        r, d = "", ""
        for i, c in enumerate(text):
            bin_c = bin(ord(c))[2:].zfill(8)
            for k in range(len(prefix), 8, 4):
                hb = bin_c[k:k+4]
                try:
                    d += code[hb]
                except KeyError:
                    d += handle_error("bcd", errors, decode=True)(hb, i)
                if len(d) == 3:
                    r += chr(int(d))
                    d = ""
        return r, len(b(text))
    return decode


add("bcd", bcd_encode(), bcd_decode(), pattern=r"^(?:bcd|binary[-_]coded[-_]decimals?)$", examples=__examples1__,
    entropy=lambda e: .45739*e+2.63519, printables_rate=.2)
add("bcd-extended0", bcd_encode("0000"), bcd_decode("0000"), examples=__examples2__, entropy=lambda e: .13584*e+2.07486,
    pattern=r"^(?:bcd|binary[-_]coded[-_]decimals?)[-_]ext(?:ended)?(?:[-_]?0|[-_]zeros?)$")
add("bcd-extended1", bcd_encode("1111"), bcd_decode("1111"), examples=__examples3__, entropy=lambda e: .13584*e+2.07486,
    pattern=r"^(?:bcd|binary[-_]coded[-_]decimals?)[-_]ext(?:ended)?(?:[-_]?1|[-_]ones?)$")




# -*- coding: UTF-8 -*-
"""Citrix Codec - citrix password encoding.

This codec:
- en/decodes strings from str to str
- en/decodes strings from bytes to bytes
- decodes file content to str (read)
- encodes file content from str to bytes (write)

Reference: https://crypto.interactive-maths.com/atbash-cipher.html
"""
from ..__common__ import *


__examples__ = {
    'enc(citrix-ctx0)':                 None,
    'enc(citrix|citrix-1|citrix_ctx1)': {'this is a test': "NBBMNAAGIDEPJJBMNIFNIMEMJKEL"},
}
__guess__ = ["citrix-ctx1"]


_dec = lambda g: ((ord(g[0]) - 0x41) & 0xf) ^ ((((ord(g[1]) - 0x41) & 0xf) << 4) & 0xf0)
_enc = lambda o: chr(((o >> 4) & 0xf) + 0x41) + chr((o & 0xf) + 0x41)


def citrix_encode(t):
    def encode(text, errors="strict"):
        l = len(text)
        r, x = "", 0
        for c in text:
            x = ord(c) ^ 0xa5 ^ x
            r += _enc(x)
        return r, l
    return encode


def citrix_decode(t):
    def decode(text, errors="strict"):
        l = len(text)
        text = text[::-1]
        r = ""
        for i in range(0, l, 2):
            x = 0 if i + 2 >= l else _dec(text[i+2:i+4])
            x ^= _dec(text[i:i+2]) ^ 0xa5
            r += chr(x)
        return r[::-1], l
    return decode


add("citrix", citrix_encode, citrix_decode, r"citrix(|[-_]?(?:ctx)?1)$", entropy=4., printables_rate=1.)




# -*- coding: UTF-8 -*-
"""Base91 Codec - base91 content encoding.

This codec:
- en/decodes strings from str to str
- en/decodes strings from bytes to bytes
- decodes file content to str (read)
- encodes file content from str to bytes (write)
"""
from ._base import digits, lower, main, upper
from ..__common__ import *

# no __examples__ ; handled manually in tests/test_base.py
__guess__ = ["base91", "base91-inv", "base91-alt", "base91-alt-inv"]


B91 = {
    '':        upper + lower + digits + "!#$%&()*+,./:;<=>?@[]^_`{|}~\"",
    'inv':     lower + upper + digits + "!#$%&()*+,./:;<=>?@[]^_`{|}~\"",
    'alt':     "!#$%&'()*+,-./" + digits + ":;<=>?@" + upper + "[\\]^_" + lower + "{|}",
    'alt-inv': "!#$%&'()*+,-./" + digits + ":;<=>?@" + lower + "[\\]^_" + upper + "{|}",
}


__chr = lambda c: chr(c) if isinstance(c, int) else c
__ord = lambda c: ord(c) if not isinstance(c, int) else c


class Base91DecodeError(ValueError):
    pass


def base91_encode(mode):
    mode = mode.replace("alternate", "alt").replace("inverted", "inv").replace("_", "-").lstrip("-")
    b91 = B91[mode if mode in B91.keys() else ""]
    def encode(text, errors="strict"):
        t, s, bits = b(text), "", ""
        if mode.startswith("alt"):
            while len(bits) < 13 and t:
                bits += "{:08b}".format(__ord(t[0]))
                t = t[1:]
            while len(bits) > 13 or t:
                n = int(bits[:13], 2)
                s += b91[n // 91] + b91[n % 91]
                bits = bits[13:]
                while len(bits) < 13 and t:
                    bits += "{:08b}".format(__ord(t[0]))
                    t = t[1:]
            if len(bits) > 0:
                if len(bits) < 7:
                    bits += "0" * (6 - len(bits))
                    s += b91[int(bits, 2)]
                else:
                    bits += "0" * (13 - len(bits))
                    n = int(bits, 2)
                    s += b91[n // 91] + b91[n % 91]
        else:
            for c in t:
                bits = bin(__ord(c))[2:].zfill(8) + bits
                if len(bits) > 13:
                    n = int(bits[-13:], 2)
                    if n > 88:
                        bits = bits[:-13]
                    else:
                        n = int(bits[-14:], 2)
                        bits = bits[:-14]
                    s += b91[n % 91] + b91[n // 91]
            if len(bits) > 0:
                n = int(bits, 2)
                s += b91[n % 91]
                if len(bits) > 7 or n > 90:
                    s += b91[n // 91]
        return s, len(t)
    return encode


def base91_decode(mode):
    mode = mode.replace("alternate", "alt").replace("inverted", "inv").replace("_", "-").lstrip("-")
    b91 = {c: i for i, c in enumerate(B91[mode if mode in B91.keys() else ""])}
    def decode(text, errors="strict"):
        t, s, bits, alt = b(text), "", "", mode.startswith("alt")
        ehandler = handle_error("base91", errors, Base91DecodeError, decode=True)
        for i in range(0, len(t), 2):
            try:
                n = b91[__chr(t[i])] * [1, 91][alt]
            except KeyError:
                ehandler(__chr(t[i]), i, s)
            try:
                j = i + 1
                n += b91[__chr(t[j])] * [91, 1][alt]
            except IndexError:
                pass
            except KeyError:
                ehandler(__chr(t[j]), j, s)
            if alt:
                bits += "{:013b}".format(n)
                while 8 <= len(bits):
                    s += chr(int(bits[0:8], 2))
                    bits = bits[8:]
            else:
                bits = bin(n)[2:].zfill([14, 13][n & 8191 > 88]) + bits
                while len(bits) > 8:
                    s += chr(int(bits[-8:], 2))
                    bits = bits[:-8]
        if alt and len(t) % 2 == 1:
            bits += "{:06b}".format(b91[__chr(t[-1])])
            while 8 <= len(bits):
                s += chr(int(bits[:8], 2))
                bits = bits[8:]
        elif not alt and len(bits) > 0 and not set(bits) == {"0"}:
            s += chr(int(bits, 2))
        return s, len(t)
    return decode


add("base91", base91_encode, base91_decode, r"^base[-_]?91((?:|[-_]alt(?:ernate)?)(?:|[-_]inv(?:erted)?)?)$",
    entropy=6.5)
main = main(91, "<http://base91.sourceforge.net/>")




# -*- coding: UTF-8 -*-
"""Rotate-Bits Codec - rotate-N-bits content encoding.

This codec:
- en/decodes strings from str to str
- en/decodes strings from bytes to bytes
- decodes file content to str (read)
- encodes file content from str to bytes (write)
"""
from ..__common__ import *


__examples__ = {
    'enc(rotate-0|rotate-8|rotate-left-8)': None,
    'enc(rotate1|rotate-right-1|rotate_1)': {'This is a test': "*4\xb4\xb9\x10\xb4\xb9\x10\xb0\x10:\xb2\xb9:"},
    'enc(rotate-left-1|rotate_left_1)':     {'This is a test': "¨ÐÒæ@Òæ@Â@èÊæè"},
}
__guess__ = ["rotate-%d" % i for i in range(1, 8)] + ["rotate-left-%d" % i for i in range(1, 8)]


if PY3:
    def _getn(i):
        m = 1
        if str(i).startswith("left"):
            i = i[4:].lstrip("-_")
            m = -1
        return m * int(i)


    def _rotaten(text, n=1):
        r = ""
        for c in ensure_str(text):
            b = bin(ord(c))[2:].zfill(8)
            r += chr(int(b[-n:] + b[:-n], 2))
        return r


    def rotate_encode(i):
        def encode(text, errors="strict"):
            return _rotaten(text, _getn(i)), len(text)
        return encode


    def rotate_decode(i):
        def decode(text, errors="strict"):
            return _rotaten(text, -_getn(i)), len(text)
        return decode


    add("rotate", rotate_encode, rotate_decode, r"rotate(?:[-_]?bits)?[-_]?((?:(?:left|right)[-_]?)?[1-7])$", entropy=1.)




#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""Automatically generated codec tests.

"""
import os
import re
from itertools import chain
from random import randint
from string import printable
from unittest import TestCase

from codext.__common__ import *


def make_test(**params):
    """ Test factory function for auto-creating tests for encodings having __examples__ defined. """
    def _template(self):
        tfile = "test-codec-%s.txt" % params['name']
        icase = params.get('ignore_case')
        icdec = lambda s: s.lower() if icase in ["decode", "both"] else s
        icenc = lambda s: s.lower() if icase in ["encode", "both"] else s
        # first, define if only encode is used ; if so, decoding must occur right after encode tests, otherwise just
        #  execute the defined decode tests
        dec = True
        for k in params['examples'].keys():
            if k.startswith("dec"):
                dec = False
        # now execute tests relying on the given examples
        for k, examples in params['examples'].items():
            # multiple encoding names can be given, e.g. 'enc(morse|morse-AB|...)'
            m = re.match(r"(?:dec|enc|enc-dec)\((.*?)(?:\|(.*?))*\)", k)
            if m:
                f1 = getattr(codecs, ["decode", "encode"][k.startswith("enc")])
                f2 = getattr(codecs, ["encode", "decode"][k.startswith("enc")])
                for ename in m.groups():
                    if ename is None:
                        continue
                    # erroneous encoding name test
                    if examples is None:
                        self.assertRaises(LookupError, f1, "test", ename)
                        continue
                    # unhandled character error tests
                    encmap = params.get('encmap')
                    if encmap and params['intype'] not in ["bin", "ord"] and not params['no_error']:
                        if not isinstance(encmap, list):
                            encmap = [encmap]
                        for em in encmap:
                            if k.startswith("dec"):
                                em = {v: k for k, v in em.items()}
                            # find one handled character and one unhandled
                            c1, c2 = None, None
                            p = list(map(ord, printable))
                            for i in chain(p, set(range(256)) - set(p)):
                                if chr(i) in em.keys():
                                    c1 = chr(i)
                                    break
                            for i in chain(set(range(256)) - set(p), p):
                                if chr(i) not in em.keys():
                                    c2 = chr(i)
                                    break
                            # now check that it raises the right error or not given the selected errors handling
                            if c1 and c2:
                                sep = params['sep'][0] if len(params['sep']) > 0 else ""
                                self.assertRaises(ValueError, f1, c2, ename)
                                self.assertRaises(ValueError, f1, c2, ename, "BAD_ERRORS")
                                if not k.startswith("enc-dec"):
                                    self.assertEqual(f1(c1 + c2, ename, "ignore"), f1(c1, ename))
                                    self.assertEqual(f1(c1 + c2, ename, "leave"), f1(c1, ename) + sep + c2)
                                    self.assertEqual(f1(c1 + c2, ename, "replace"), f1(c1, ename) + sep + \
                                                     params.get('repl_minlen', 1) * params['repl_char'])
                    # examples validation tests
                    if k.startswith("enc-dec") and isinstance(examples, list):
                        for e in examples[:]:
                            rd = re.match(r"\@random(?:\{(\d+(?:,(\d+))*?)\})?$", e)
                            if rd:
                                examples.remove(e)
                                for n in (rd.group(1) or "512").split(","):
                                    examples.append("".join(chr(randint(0, 255)) for i in range(int(n))))
                        for s in [""] + examples:
                            self.assertEqual(icdec(f2(icenc(f1(s, ename)), ename)), icdec(s))
                            self.assertEqual(icdec(f2(icenc(f1(b(s), ename)), ename)), b(icdec(s)))
                            # file tests
                            with codecs.open(tfile, 'wb', encoding=ename) as f:
                                f.write(b(s))
                            with codecs.open(tfile, 'rb', encoding=ename) as f:
                                s2 = f.read() if PY3 else f.read().rstrip("\x00")
                            self.assertEqual(b(icdec(s2)), b(icdec(s)))
                            os.remove(tfile)
                    else:
                        for s1, s2 in examples.items():
                            # willingly erroneous tests
                            if s2 is None:
                                self.assertRaises(ValueError, f1, s1, ename)
                                continue
                            # raw text tests
                            self.assertEqual(icenc(f1(s1, ename)), icenc(s2))
                            self.assertEqual(b(icenc(f1(s1, ename))), b(icenc(s2)))
                            self.assertIsNotNone(f1(s1, ename, "replace"))
                            self.assertIsNotNone(f1(s1, ename, "ignore"))
                            if dec:
                                self.assertEqual(icdec(f2(s2, ename)), icdec(s1))
                                self.assertEqual(b(icdec(f2(s2, ename))), b(icdec(s1)))
                                self.assertIsNotNone(f2(s2, ename, "replace"))
                                self.assertIsNotNone(f2(s2, ename, "ignore"))
                            if k.startswith("enc"):
                                # file tests
                                with codecs.open(tfile, 'wb', encoding=ename) as f:
                                    f.write(b(s1))
                                with codecs.open(tfile, 'rb', encoding=ename) as f:
                                    s = f.read() if PY3 else f.read().rstrip("\x00")
                                self.assertEqual(b(icdec(f2(s2, ename))), b(icdec(s)))
                                os.remove(tfile)
    return _template


class GeneratedTestCase(TestCase):
    pass


for encoding in list_encodings():
    try:
        ci = lookup(encoding)
    except LookupError:
        continue
    # only consider codecs with __examples__ defined in their globals for dynamic tests generation
    if ci.parameters.get('examples') is not None:
        f = make_test(**ci.parameters)
        f.__name__ = n = "test_" + encoding.replace("-", "_")
        setattr(GeneratedTestCase, n, f)




# -*- coding: UTF-8 -*-
"""LZ77 Codec - Lempel-Ziv 1977 compression algorithm.

NB: Not an encoding properly speaking.

This codec:
- en/decodes strings from str to str
- en/decodes strings from bytes to bytes
- decodes file content to str (read)
- encodes file content from str to bytes (write)

Inspired from: https://github.com/manassra/LZ77-Compressor
"""
from ..__common__ import *


__examples__ = {'enc-dec(lz77)': ["test", "This is a test", "@random{512,1024,2048}"]}


_B2b = lambda B: bin(B if isinstance(B, int) else ord(B))[2:].zfill(8)
_b2B = lambda bt: "".join(chr(int(bt[i:i+8], 2)) for i in range(0, len(bt), 8))
WINDOW_SIZE = 20


def _find_longest_match(data, pos):
    """ Finds the longest match to a substring starting at the current position (pos) in the lookahead buffer from
         the history window. """
    eob, bmd, bml = min(pos + 15, len(data) + 1), -1, -1
    for j in range(pos + 2, eob):
        start = max(0, pos - WINDOW_SIZE)
        substr = data[pos:j]
        l = len(substr)
        for i in range(start, pos):
            n, r = l // (pos - i), l % (pos - i)
            if data[i:pos] * n + data[i:i+r] == substr and l > bml:
                bmd, bml = pos - i, l
    if bmd > 0 and bml > 0:
        return bmd, bml


def lz77_compress(input, errors="strict"):
    """ Compresses the given data by applying LZ77 compression algorithm. """
    i, l, bits = 0, len(input), ""
    while i < l:
        try:
            bmd, bml = _find_longest_match(input, i)
            bits += "1" + _B2b(bmd >> 4) + _B2b(((bmd & 0xf) << 4) | bml)
            i += bml
        except TypeError:
            bits += "0" + _B2b(input[i])
            i += 1
    bits += "0" * ((8 - (len(bits) % 8)) % 8)
    return _b2B(bits), l


def lz77_decompress(input, errors="strict"):
    """ Decompresses the given data. """
    out, d = "", "".join(_B2b(c) for c in input)
    while len(d) >= 9:
        flag, d = d[0], d[1:]
        if flag == "0":
            out += _b2B(d[:8])
            d = d[8:]
        else:
            B1, B2 = int(d[:8], 2), int(d[8:16], 2)
            d = d[16:]
            dist = (B1 << 4) | (B2 >> 4)
            for i in range(B2 & 0xf):
                out += out[-dist]
    return out, len(out)


add("lz77", lz77_compress, lz77_decompress, entropy=7.9)




# -*- coding: UTF-8 -*-
"""Gzip Codec - gzip content compression.

NB: Not an encoding properly speaking.

This codec:
- en/decodes strings from str to str
- en/decodes strings from bytes to bytes
- decodes file content to str (read)
- encodes file content from str to bytes (write)
"""
import zlib
from gzip import GzipFile

from ..__common__ import *


__examples__ = {'enc-dec(gzip)': ["test", "This is a test", "@random{512,1024,2048}"]}


def gzip_compress(text, errors="strict"):
    out = BytesIO()
    with GzipFile(fileobj=out, mode="wb") as f:
        f.write(b(text))
    return out.getvalue(), len(text)


def gzip_decompress(data, errors="strict"):
    # then try decompressing considering the file signature
    try:
        with GzipFile(fileobj=BytesIO(b(data)), mode="rb") as f:
            r = f.read()
    except:
        pass
    # try decompressing without considering the file signature
    try:
        r = zlib.decompress(b(data), 16 + zlib.MAX_WBITS)
    except:
        return handle_error("gzip", errors, decode=True)(data[0], 0) if len(data) > 0 else "", len(data)
    return r, len(r)


add("gzip", gzip_compress, gzip_decompress, entropy=7.9)




# -*- coding: UTF-8 -*-
"""Shift Codec - Shift-ordinal-with-N content encoding.

This codec:
- en/decodes strings from str to str
- en/decodes strings from bytes to bytes
- decodes file content to str (read)
- encodes file content from str to bytes (write)
"""
from ..__common__ import *


__examples__ = {
    'enc(shift0|shift--10|shift256)': None,
    'enc(shift1|shift_1|shift-1)':    {'this is a test': "uijt!jt!b!uftu"},
    'enc(shift9|shift_9|shift-9)':    {'this is a test': "}qr|)r|)j)}n|}"},
}
__guess__ = ["shift-%d" % i for i in range(1, 256)]


def ord_shift_decode(i):
    return ord_shift_encode(-i)


def ord_shift_encode(i):
    def encode(text, errors="strict"):
        r = "".join(chr((ord(c) + i) % 256) for c in text)
        return r, len(r)
    return encode


add("shift", ord_shift_encode, ord_shift_decode, r"shift[-_]?([1-9]|[1-9][0-9]|1[0-9][0-9]|2[0-4][0-9]|25[0-5])$")




# -*- coding: UTF-8 -*-
"""Radio Codec - NATO/Military phonetic alphabet content encoding.

This codec:
- en/decodes strings from str to str
- en/decodes strings from bytes to bytes
- decodes file content to str (read)
- encodes file content from str to bytes (write)
"""
from ..__common__ import *


__examples__ = {
    'enc(radio|military-alphabet)':      {'test':   "Tango Echo Sierra Tango"},
    'enc(nato-alphabet|radio-phonetic)': {'string': "Sierra Tango Romeo India November Golf"},
}


ENCMAP = {
    'A': "Alpha", 'B': "Bravo", 'C': "Charlie", 'D': "Delta", 'E': "Echo", 'F': "Foxtrot", 'G': "Golf", 'H': "Hotel",
    'I': "India", 'J': "Juliett", 'K': "Kilo", 'L': "Lima", 'M': "Mike", 'N': "November", 'O': "Oscar", 'P': "Papa",
    'Q': "Quebec", 'R': "Romeo", 'S': "Sierra", 'T': "Tango", 'U': "Uniform", 'V': "Victor", 'W': "Whiskey",
    'X': "X-ray", 'Y': "Yankee", 'Z': "Zulu", ' ': "/",
}


add_map("radio", ENCMAP, sep=" ", ignore_case="both",
        pattern=r"^(?:military|nato|radio)(?:(?:[-_]phonetic)?(?:[-_]alphabet)?)?$", printables_rate=1.)




# -*- coding: UTF-8 -*-
"""Base122 Codec - base122 content encoding.

This codec:
- en/decodes strings from str to str
- en/decodes strings from bytes to bytes
- decodes file content to str (read)
- encodes file content from str to bytes (write)
"""
from ._base import main
from ..__common__ import *


__examples__ = {
    'enc(base122|base-122)': {
        'this is a test': ":\x1aʗ\x19\x01Rs\x10\x18$\x07#\x15ft",
        b'This is another longer test string with d1g1t5 and sp3c141 characters !\n': \
            b"*\x1a\xca\x97\x19\x01Rs\x10\x18-f{QPe9\x08\xcb\x86{9Ne9\x08\x0eF+Mh 9]\x0e\xd3\x8b"
            b"9N ;Z.FA\x01H13L.C)\x01Bn2\x08\x0e7\x01MF1\x1a\x0c$\x06\x1b!Br0XnF+If \x10B@"
    },
    'enc-dec(base_122)': ["@random"],
}


_BAD = [0, 10, 13, 34, 38, 92]
_i = lambda c: c if isinstance(c, int) else ord(c)


def base122_encode(input, errors='strict'):
    raise NotImplementedError


def base122_decode(input, errors='strict'):
    raise NotImplementedError


if PY3:
    # inspired from: https://github.com/kevinAlbs/Base122/blob/master/base122.js
    def base122_encode(input, errors="strict"):
        idx, bit, r, l = 0, 0, [], len(input)
        
        def _get_7bits(idx, bit):
            if idx >= l:
                return idx, bit, False
            B1 = _i(input[idx])
            p1 = (((254 >> bit) & B1) << bit) >> 1
            bit += 7
            if bit < 8:
                return idx, bit, p1
            bit -= 8
            idx += 1
            if idx >= l:
                return idx, bit, p1
            B2 = _i(input[idx])
            p2 = (((65280 >> bit) & B2) & 255) >> (8 - bit)
            return idx, bit, (p1 | p2)
        
        while True:
            if idx >= l:
                break
            # get seven bits of input data
            idx, bit, B = _get_7bits(idx, bit)
            # check for illegal chars
            try:
                bad_idx = _BAD.index(B)
            except ValueError:
                r.append(B)
                continue
            idx, bit, nB = _get_7bits(idx, bit)
            if nB is False:
                nB, bad_idx = B, 7
            B1, B2 = 194, 128
            B1 |= (7 & bad_idx) << 2
            B1 |= int((nB & 64) > 0)
            B2 |= nB & 63
            r.extend([B1, B2])
        return "".join(map(chr, r)).encode("latin-1"), len(input)

    # inspired from: https://github.com/kevinAlbs/Base122/blob/master/base122.js
    def base122_decode(input, errors="strict"):
        currB, bob, r, input = 0, 0, [], list(map(ord, input))
        
        def _get_7bits(currB, bob, B, decoded):
            B <<= 1
            currB |= (B % 0x100000000) >> bob
            bob += 7
            if bob >= 8:
                decoded += [currB]
                bob -= 8
            return (B << (7 - bob)) & 255, bob

        for i in range(len(input)):
            if input[i] >= 128:
                try:
                    currB, bob = _get_7bits(currB, bob, _BAD[(input[i] >> 8) & 7], r)
                except IndexError:
                    pass
                currB, bob = _get_7bits(currB, bob, input[i] & 127, r)
            else:
                currB, bob = _get_7bits(currB, bob, input[i], r)
        return "".join(map(chr, r)), len(input)


add("base122", base122_encode, base122_decode, r"^base[-_]?122$")
main = main(122, "<http://blog.kevinalbs.com/base122>")




# -*- coding: UTF-8 -*-
"""A1Z26 Codec - A1Z26 content encoding.

This codec:
- en/decodes strings from str to str
- en/decodes strings from bytes to bytes
- decodes file content to str (read)
- encodes file content from str to bytes (write)
"""
from string import ascii_lowercase as lower

from ..__common__ import *


SEP = "-_/|,;:*"

__examples__ = {
    'enc(a1z26-BAD)': None,
    'dec(a1z26)':     {'1-12-123':       None},
    'enc(a1z26)':     {'test123': None, 'this is a test': "20-8-9-19 9-19 1 20-5-19-20"},
    'enc(a1z26-/)':   {'this is a test': "20/8/9/19 9/19 1 20/5/19/20"},
}
__guess__ = ["a1z26", "a1z26_"] + ["a1z26-" + s for s in SEP[2:]]


def a1z26_encode(sep):
    sep = sep[-1] if len(sep) > 0 else "-"
    def encode(text, errors="strict"):
        words = []
        for word in text.split():
            w = []
            for k, c in enumerate(word):
                try:
                    w.append(str(lower.index(c.lower()) + 1))
                except ValueError:
                    w.append(handle_error("a1z26", errors)(c, k))
            words.append(sep.join(w).strip(sep))
        return " ".join(words), len(text)
    return encode


def a1z26_decode(sep):
    sep = sep[-1] if len(sep) > 0 else "-"
    def decode(text, errors="strict"):
        k, words = 0, []
        for word in text.split():
            w = ""
            for i in word.split(sep):
                k += 1
                try:
                    w += lower[int(i)-1]
                except IndexError:
                    w += handle_error("a1z26", errors, decode=True)(str(i), k)
            words.append(w)
        return " ".join(words), len(text)
    return decode


add("a1z26", a1z26_encode, a1z26_decode, pattern=r"^a1z26(|[-_]|[-_][/|,;:\*])$", printables_rate=1.)




#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""Base codecs tests.

"""
import os
import sys
from unittest import TestCase

from codext.__common__ import *
from codext.base._base import _generate_charset
from codext.base.baseN import base, main2, main32, main64url


class TestCodecsBase(TestCase):
    def setUp(self):
        global STR
        STR = "this is a test"
    
    def test_new_base_codec(self):
        for i in [0, 1, 256]:
            self.assertRaises(ValueError, _generate_charset, i)
        b10 = lambda *a: "0123456789"
        base(b10, "base10")
        B10 = "2361031878030638688519054699098996"
        self.assertEqual(codecs.encode(STR, "base10"), B10)
        self.assertEqual(codecs.encode(b(STR), "base10"), b(B10))
        self.assertEqual(codecs.decode(B10, "base10"), STR)
        self.assertEqual(codecs.decode(b(B10), "base10"), b(STR))
        self.assertRaises(ValueError, base, 1, "test")
        b11 = "0123456789a"
        base(b11, "base11")
        B11 = "113342054335735319526632a26972419"
        self.assertEqual(codecs.encode(STR, "base11"), B11)
        self.assertEqual(codecs.decode(B11, "base11"), STR)
        self.assertRaises(ValueError, base, object(), "test")
        self.assertIsNone(base({'': "01234"}, r"^base5(test)?$"))
        self.assertIsNotNone(codecs.encode(STR, "base5test"))
        self.assertRaises(ValueError, base, {'': "01234"}, "base5-test", pow2=True)
    
    def test_codec_base2(self):
        STR = "test"
        B2 = "01110100011001010111001101110100"
        self.assertEqual(codecs.encode(STR, "base2"), B2)
        self.assertEqual(codecs.encode(b(STR), "base2"), b(B2))
        self.assertEqual(codecs.decode(B2, "base2"), STR)
        self.assertEqual(codecs.decode(b(B2), "base2"), b(STR))
        B2 = "10001011100110101000110010001011"
        self.assertEqual(codecs.encode(STR, "base2-inv"), B2)
        self.assertEqual(codecs.decode(B2, "base2-inv"), STR)
        B2 = "abbbabaaabbaabababbbaabbabbbabaa"
        self.assertEqual(codecs.encode(STR, "base2-ab"), B2)
        self.assertEqual(codecs.decode(B2, "base2-ab"), STR)
        B2 = "CDDDCDCCCDDCCDCDCDDDCCDDCDDDCDCC"
        self.assertEqual(codecs.encode(STR, "base2-CD"), B2)
        self.assertEqual(codecs.decode(B2, "base2-CD"), STR)
        B2 = "34443433344334343444334434443433"
        self.assertEqual(codecs.encode(STR, "base2-34"), B2)
        self.assertEqual(codecs.decode(B2, "base2-34"), STR)
    
    def test_codec_base3(self):
        STR = "test"
        B3 = "23112113223321323322"
        self.assertEqual(codecs.encode(STR, "base3"), B3)
        self.assertEqual(codecs.encode(b(STR), "base3"), b(B3))
        self.assertEqual(codecs.decode(B3, "base3"), STR)
        self.assertEqual(codecs.decode(b(B3), "base3"), b(STR))
        B3 = "21332331221123121122"
        self.assertEqual(codecs.encode(STR, "base3-inv"), B3)
        self.assertEqual(codecs.decode(B3, "base3-inv"), STR)
        B3 = "bcaabaacbbccbacbccbb"
        self.assertEqual(codecs.encode(STR, "base3-abc"), B3)
        self.assertEqual(codecs.decode(B3, "base3-abc"), STR)
        self.assertRaises(LookupError, codecs.encode, "test", "base3-ab")
        self.assertRaises(LookupError, codecs.encode, "test", "base3-abcd")
    
    def test_codec_base4(self):
        STR = "test"
        B4 = "2421232224142421"
        self.assertEqual(codecs.encode(STR, "base4"), B4)
        self.assertEqual(codecs.encode(b(STR), "base4"), b(B4))
        self.assertEqual(codecs.decode(B4, "base4"), STR)
        self.assertEqual(codecs.decode(b(B4), "base4"), b(STR))
        B4 = "3134323331413134"
        self.assertEqual(codecs.encode(STR, "base4-inv"), B4)
        self.assertEqual(codecs.decode(B4, "base4-inv"), STR)
        B4 = "bdbabcbbbdadbdba"
        self.assertEqual(codecs.encode(STR, "base4-abcd"), B4)
        self.assertEqual(codecs.decode(B4, "base4-abcd"), STR)
        self.assertRaises(LookupError, codecs.encode, "test", "base4-abc")
        self.assertRaises(LookupError, codecs.encode, "test", "base4-abcde")
    
    def test_codec_base8(self):
        STR = "test"
        B8 = "dfagcfgddfa====="
        self.assertEqual(codecs.encode(STR, "base8"), B8)
        self.assertEqual(codecs.encode(b(STR), "base8"), b(B8))
        self.assertEqual(codecs.decode(B8, "base8"), STR)
        self.assertEqual(codecs.decode(b(B8), "base8"), b(STR))
        B8 = "echbfcbeech====="
        self.assertEqual(codecs.encode(STR, "base8-inv"), B8)
        self.assertEqual(codecs.decode(B8, "base8-inv"), STR)
        B8 = "35062563350====="
        self.assertEqual(codecs.encode(STR, "base8-01234567"), B8)
        self.assertEqual(codecs.decode(B8, "base8-01234567"), STR)
        self.assertRaises(LookupError, codecs.encode, "test", "base8-0123456")
        self.assertRaises(LookupError, codecs.encode, "test", "base8-012345678")
    
    def test_codec_base16(self):
        B16 = "7468697320697320612074657374"
        self.assertEqual(codecs.encode(STR, "base16"), B16)
        self.assertEqual(codecs.encode(b(STR), "base16"), b(B16))
        self.assertEqual(codecs.decode(B16, "base16"), STR)
        self.assertEqual(codecs.decode(b(B16), "base16"), b(STR))
        B16 += "?"
        self.assertRaises(ValueError, codecs.decode, B16, "base16")
        self.assertEqual(codecs.decode(B16, "base16", "ignore"), STR)
        self.assertEqual(codecs.decode(B16, "base16", "replace"), STR + "\x00")
        self.assertRaises(ValueError, codecs.decode, B16, "base16", "BAD")
        STR2 = "=:;"
        B16_1 = "3d3a3b"
        B16_2 = "3D3A3B"
        B16_3 = "3D3a3B"  # mixed case: should fail
        self.assertEqual(codecs.encode(STR2, "hex"), B16_2)
        self.assertEqual(codecs.decode(B16_1, "hex"), STR2)
        self.assertEqual(codecs.decode(B16_2, "hex"), STR2)
        self.assertRaises(ValueError, codecs.decode, B16_3, "hex")
    
    def test_codec_base32(self):
        B32 = "ORUGS4ZANFZSAYJAORSXG5A="
        self.assertEqual(codecs.encode(STR, "base32"), B32)
        self.assertEqual(codecs.encode(b(STR), "base32"), b(B32))
        self.assertEqual(codecs.decode(B32, "base32"), STR)
        self.assertEqual(codecs.decode(b(B32), "base32"), b(STR))
        B32 = "qtwg1h3ypf31yajyqt1zg7y="
        self.assertEqual(codecs.encode(STR, "zbase32"), B32)
        self.assertEqual(codecs.encode(b(STR), "z-base-32"), b(B32))
        self.assertEqual(codecs.decode(B32, "z_base_32"), STR)
        self.assertEqual(codecs.decode(b(B32), "zbase32"), b(STR))
        self.assertRaises(ValueError, codecs.decode, B32.rstrip("="), "zbase32")
        self.assertRaises(ValueError, codecs.decode, B32.rstrip("="), "zbase32", "BAD")
    
    def test_codec_base36(self):
        B36 = "4WMHTK6UZL044O91NKCEB8"
        self.assertEqual(codecs.encode(STR, "base36"), B36)
        self.assertEqual(codecs.encode(b(STR), "base36"), b(B36))
        self.assertEqual(codecs.decode(B36, "base36"), STR)
        self.assertEqual(codecs.decode(b(B36), "base36"), b(STR))
        B36 = "E6WR3UG49VAEEYJBXUMOLI"
        self.assertEqual(codecs.encode(STR, "base36-inv"), B36)
        self.assertEqual(codecs.decode(B36, "base36-inv"), STR)
        self.assertRaises(ValueError, codecs.decode, B36 + "?", "base36-inv")
        self.assertRaises(ValueError, codecs.decode, B36 + "?", "base36", "BAD")
        self.assertEqual(codecs.decode(B36 + "?", "base36-inv", "ignore"), STR)
    
    def test_codec_base58(self):
        B58 = "jo91waLQA1NNeBmZKUF"
        self.assertEqual(codecs.encode(STR, "base58"), B58)
        self.assertEqual(codecs.encode(b(STR), "base58"), b(B58))
        self.assertEqual(codecs.decode(B58, "base58"), STR)
        self.assertEqual(codecs.decode(b(B58), "base58"), b(STR))
        B58 = "jo9rA2LQwr44eBmZK7E"
        self.assertEqual(codecs.encode(STR, "base58-ripple"), B58)
        self.assertEqual(codecs.decode(B58, "base58-rp"), STR)
        B58 = "JN91Wzkpa1nnDbLyjtf"
        self.assertEqual(codecs.encode(STR, "base58-flickr"), B58)
        self.assertEqual(codecs.encode(STR, "base58-shorturl"), B58)
        self.assertEqual(codecs.decode(B58, "base58-fl"), STR)
        self.assertEqual(codecs.encode(STR, "base58-short-url"), B58)
        self.assertEqual(codecs.encode(STR, "base58-url"), B58)
    
    def test_codec_base62(self):
        B62 = "CsoB4HQ5gmgMyCenF7E"
        self.assertEqual(codecs.encode(STR, "base62"), B62)
        self.assertEqual(codecs.encode(b(STR), "base62"), b(B62))
        self.assertEqual(codecs.decode(B62, "base62"), STR)
        self.assertEqual(codecs.decode(b(B62), "base62"), b(STR))
        B62 = "cSOb4hq5GMGmYcENf7e"
        self.assertEqual(codecs.encode(STR, "base62-inv"), B62)
        self.assertEqual(codecs.decode(B62, "base62-inv"), STR)
    
    def test_codec_base64(self):
        B64 = "dGhpcyBpcyBhIHRlc3Q="
        self.assertEqual(codecs.encode(STR, "base64"), B64)
        self.assertEqual(codecs.encode(b(STR), "base64"), b(B64))
        self.assertEqual(codecs.decode(B64, "base64"), STR)
        self.assertEqual(codecs.decode(b(B64), "base64"), b(STR))
        B64 = "DgHPCYbPCYbHihrLC3q="
        self.assertEqual(codecs.encode(STR, "base64-inv"), B64)
        self.assertEqual(codecs.decode(B64, "base64-inv"), STR)
    
    def test_codec_base85(self):
        if PY3:
            B85 = "bZBXFAZc?TVIXv6b94"
            self.assertEqual(codecs.encode(STR, "base85"), B85)
            self.assertEqual(codecs.encode(b(STR), "base85"), b(B85))
            self.assertEqual(codecs.decode(B85, "base85"), STR)
            self.assertEqual(codecs.decode(b(B85), "base85"), b(STR))
    
    def test_codec_base91(self):
        B91 = ",X,<:WRT%yxth90oZB"
        self.assertEqual(codecs.encode(STR, "base91"), B91)
        self.assertEqual(codecs.encode(b(STR), "base91"), b(B91))
        self.assertEqual(codecs.decode(B91, "base91"), STR)
        self.assertEqual(codecs.decode(b(B91), "base91"), b(STR))
        B91 = ",x,<:wrt%YXTH90Ozb"
        self.assertEqual(codecs.encode(STR, "base91-inv"), B91)
        self.assertEqual(codecs.decode(B91, "base91-inv"), STR)
        self.assertIsNotNone(codecs.encode("\x00\x00", "base91"))
        self.assertIsNotNone(codecs.decode("abc", "base91"))
        self.assertIsNotNone(codecs.decode("AD", "base91"))
        self.assertRaises(ValueError, codecs.decode, "\xff", "base91")
        self.assertRaises(ValueError, codecs.decode, "a\xff", "base91")
        B91A = "Jx&[jv4S3Wg>,71@Jk"
        self.assertEqual(codecs.encode(STR, "base91-alt"), B91A)
        self.assertEqual(codecs.encode(b(STR), "base91-alt"), b(B91A))
        self.assertEqual(codecs.decode(B91A, "base91_alt"), STR)
        self.assertEqual(codecs.decode(b(B91A), "base91_alt"), b(STR))
        B91A = "jX&[JV4s3wG>,71@jK"
        self.assertEqual(codecs.encode(STR, "base91-alt-inv"), B91A)
        self.assertEqual(codecs.decode(B91A, "base91_alt_inv"), STR)
        self.assertIsNotNone(codecs.encode("\x00\x00", "base91-alt"))
        self.assertIsNotNone(codecs.decode("abc", "base91-alt"))
        self.assertIsNotNone(codecs.decode("AD", "base91_alt"))
    
    def test_codec_base100(self):
        if PY3:
            B100 = "\U0001f46b\U0001f45f\U0001f460\U0001f46a\U0001f417\U0001f460\U0001f46a\U0001f417\U0001f458" \
                   "\U0001f417\U0001f46b\U0001f45c\U0001f46a\U0001f46b"
            self.assertEqual(codecs.encode(STR, "base100"), B100)
            self.assertEqual(codecs.encode(b(STR), "base100"), b(B100))
            self.assertEqual(codecs.decode(B100, "base100"), STR)
            self.assertEqual(codecs.decode(b(B100), "base100"), b(STR))
            self.assertRaises(ValueError, codecs.decode, b(B100)[1:], "base100")
    
    def test_codec_base_generic(self):
        for n in range(2, 255):
            bn = "base{}_generic".format(n)
            self.assertEqual(codecs.decode(codecs.encode(STR, bn), bn), STR)
        self.assertRaises(LookupError, codecs.decode, "test", "base0-generic")
        self.assertRaises(LookupError, codecs.decode, "test", "base1-generic")
        self.assertRaises(LookupError, codecs.decode, "test", "base256-generic")
    
    def test_base_main(self):
        tmp = sys.argv[:]
        tfile = "test-base-main.txt"
        with open(tfile, 'w') as f:
            f.write("This is a long test string for the sake of causing line wrapping based on default parameters.")
        sys.argv = [tmp[0], tfile]
        for m in main32, main64url:
            self.assertEqual(m(), 0)
        sys.argv = [tmp[0], tfile, "-d"]
        self.assertEqual(main2(), 1)
        os.remove(tfile)
        sys.argv[:] = tmp




# -*- coding: UTF-8 -*-
"""XOR Codec - XOR-with-1-byte content encoding.

This codec:
- en/decodes strings from str to str
- en/decodes strings from bytes to bytes
- decodes file content to str (read)
- encodes file content from str to bytes (write)
"""
from ..__common__ import *


__examples__ = {
    'enc(xor0|xor--10|xor256|xor300)': None,
    'enc(xor3|xor-3|xor_3)':           {'this is a test': "wkjp#jp#b#wfpw"},
    'enc(xor3|xor-3|xor_3)':           {'wkjp#jp#b#wfpw': "this is a test"},
    'enc(xor6|xor-6|xor_6)':           {'this is a test': "rnou&ou&g&rcur"},
}
__guess__ = ["xor-%d" % i for i in range(1, 256)]


def _xorn(text, n=1):
    return "".join(chr(ord(c) ^ (n % 256)) for c in text)


def xor_byte_encode(i):
    def encode(text, errors="strict"):
        r = _xorn(ensure_str(text), i)
        return r, len(r)
    return encode


add("xor", xor_byte_encode, xor_byte_encode, r"^xor[-_]?([1-9]|[1-9][0-9]|1[0-9][0-9]|2[0-4][0-9]|25[0-5])$")




# -*- coding: UTF-8 -*-
"""Excess-3 Codec - Excess-3 code (aka Stibitz code) content encoding.

This codec:
- en/decodes strings from str to str
- en/decodes strings from bytes to bytes
- decodes file content to str (read)
- encodes file content from str to bytes (write)
"""
from ..__common__ import *


__examples__ = {
    'enc(excess3|xs-3|stibitz)': {
        'This is a test!':       ";t7C\x84H6T8D\x83e<\xa3eD\x944D\x84I6`",
        'This is another test ': ";t7C\x84H6T8D\x83e<\xa4CDDICt4DseD\x944D\x84I6P",
    },
    'dec(excess-3|xs3)': {
        '\x00':                                            None,
        '\xff':                                            None,
        ';t7C\x84H6T8D\x83e<\xa3eD\x944D\x84I6`':          "This is a test!",
        ';t7C\x84H6T8D\x83e<\xa4CDDICt4DseD\x944D\x84I6P': "This is another test ",
    },
}


CODE = {
    '0': "0011", '1': "0100", '2': "0101", '3': "0110", '4': "0111",
    '5': "1000", '6': "1001", '7': "1010", '8': "1011", '9': "1100",
}


def excess3_encode(text, errors="strict"):
    r, bits = "", ""
    for c in text:
        for i in str(ord(c)).zfill(3):
            bits += CODE[i]
            if len(bits) == 8:
                r += chr(int(bits, 2))
                bits = ""
    if len(bits) > 0:
        r += chr(int(bits + "0000", 2))
    return r, len(b(text))


def excess3_decode(text, errors="strict"):
    code = {v: k for k, v in CODE.items()}
    r, d = "", ""
    for i, c in enumerate(text):
        bin_c = bin(ord(c))[2:].zfill(8)
        for k in range(0, 8, 4):
            hb = bin_c[k:k+4]
            try:
                d += code[hb]
            except KeyError:  # (normal case) occurs when 0000 was used for padding
                if i != len(text) - 1 or k != 4 or hb != "0000":
                    d += handle_error("excess3", errors, decode=True)(hb, i)
            if len(d) == 3:
                r += chr(int(d))
                d = ""
    return r, len(b(text))


add("excess3", excess3_encode, excess3_decode, pattern=r"^(?:excess\-?3|xs\-?3|stibitz)$", entropy=1.,
    printables_rate=.45)




# -*- coding: UTF-8 -*-
"""BaseN Codecs - base content encodings.

These codecs:
- en/decode strings from str to str
- en/decode strings from bytes to bytes
- decode file content to str (read)
- encode file content from str to bytes (write)
"""
from ..__common__ import *
from ._base import base, base_generic, digits, lower, main, upper
from ._base2n import base2n


B2 = {r'': "01", r'[-_]inv(erted)?': "10"}
base2n(B2, r"^(?:base[-_]?2|bin)(|[-_]inv(?:erted)?|[-_][a-zA-Z0-9]{2})$")
main2 = main(2)


B3 = {r'': "123", r'[-_]inv(erted)?': "321"}
base(B3, r"^base[-_]?3(|[-_]inv(?:erted)?|[-_][a-zA-Z0-9]{3})$")
main3 = main(3)


B4 = {r'': "1234", r'[-_]inv(erted)?': "4321"}
base2n(B4, r"^base[-_]?4(|[-_]inv(?:erted)?|[-_][a-zA-Z0-9]{4})$")
main4 = main(4)


B8 = {r'': "abcdefgh", r'[-_]inv(erted)?': "hgfedcba"}
base2n(B8, r"^base[-_]?8(|[-_]inv(?:erted)?|[-_][a-zA-Z0-9]{8})$")
main8 = main(8)


B16 = {'': digits + "ABCDEF", 'inv': "ABCDEF" + digits}
base2n(B16, r"^(?:base[-_]?16|hex)(|[-_]inv(?:erted)?)$")
main16 = main(16, "RFC 4648")


B32 = {
    r'':                               upper + "234567",
    r'[-_]inv(erted)?$':               "234567" + upper,
    r'(?:[-_]ext(?:ended)?)?[-_]hex$': digits + upper[:22],
    r'[-_]geohash':                    digits + "bcdefghjkmnpqrstuvwxyz",
}
base2n(B32, r"^base[-_]?32(|[-_]inv(?:erted)?|(?:[-_]ext(?:ended)?)?[-_]hex|[-_]geohash)$", padding_char="=",
       guess=["base32", "base32-inv", "base32-hex", "base32-geohash"])
main32 = main(32, "RFC 4648")
main32hex = main(32, "RFC 4648", "hex", False)
main32geo = main(32, "RFC 4648", "geohash", False)


ZB32 = {'': "ybndrfg8ejkmcpqxot1uwisza345h769"}
base2n(ZB32, r"^z[-_]?base[-_]?32$", name="zbase32", padding_char="=")
mainz32 = main(32, "<https://philzimmermann.com/docs/human-oriented-base-32-encoding.txt>", "z", False)


B36 = {'': digits + upper, 'inv': upper + digits}
base(B36, r"^base[-_]?36(|[-_]inv(?:erted)?)$")
main36 = main(36, "<https://en.wikipedia.org/wiki/Base36>")


B58 = {
    r'(|[-_](bc|bitcoin))$':              "123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz",
    r'[-_](rp|ripple)$':                  "rpshnaf39wBUDNEGHJKLM4PQRST7VWXYZ2bcdeCg65jkm8oFqi1tuvAxyz",
    r'[-_](fl|flickr|short[-]?url|url)$': "123456789abcdefghijkmnopqrstuvwxyzABCDEFGHJKLMNPQRSTUVWXYZ",
}
base(B58, r"^base[-_]?58(|[-_](bc|bitcoin|rp|ripple|fl|flickr|short[-]?url|url))$",
     guess=["base58-bitcoin", "base58-ripple", "base58-flickr"])
main58bc = main(58, "<https://en.bitcoinwiki.org/wiki/Base58>", "bitcoin")
main58rp = main(58, "<https://en.bitcoinwiki.org/wiki/Base58>", "ripple")
main58fl = main(58, "<https://en.bitcoinwiki.org/wiki/Base58>", "flickr")


B62 = {'': digits + upper + lower, 'inv': digits + lower + upper}
base(B62, r"^base[-_]?62(|[-_]inv(?:erted)?)$")
main62 = main(62, "<https://en.wikipedia.org/wiki/Base62>")


B63 = {'': upper + lower + digits + "$", 'inv': lower + upper + digits + "$"}
base(B63, r"^base[-_]?63(|[-_]inv(?:erted)?)$")
main63 = main(63)


B64 = {
    r'':                        upper + lower + digits + "+/",
    r'[-_]inv(erted)?$':        lower + upper + digits + "+/",
    r'[-_]?(file|url)(safe)?$': upper + lower + digits + "-_",
}
base2n(B64, r"^base[-_]?64(|[-_]inv(?:erted)?|[-_]?(?:file|url)(?:safe)?)$", padding_char="=",
       guess=["base64", "base64-inv", "base64-url"])
main64 = main(64, "RFC 4648")
main64url = main(64, "RFC 4648 / Base64URL", "url", False)


B67 = {
    r'':                 upper + lower + digits + "-_.!~",
    r'[-_]inv(erted)?$': lower + upper + digits + "-_.!~",
}
base(B67, r"^base[-_]?67(|[-_]inv(?:erted)?)$")
main67 = main(67)


# generic base encodings, to be added after all others as they have the precedence
base_generic()




# -*- coding: UTF-8 -*-
"""Leetspeak Codec - leetspeak content encoding.

This codec:
- en/decodes strings from str to str
- en/decodes strings from bytes to bytes
- decodes file content to str (read)
- encodes file content from str to bytes (write)
"""
from ..__common__ import *


__examples__ = {
    'enc(leet|1337|leetspeak)': {'this is a test': "7h15 15 4 7357"},
    'dec(leet|1337|leetspeak)': {'7H15 15 4 7357': "THIS IS A TEST"},
}


ENCMAP = {k: v for k, v in zip("abeiostz", "48310572")}


add_map("leet", ENCMAP, ignore_case="encode", no_error=True, pattern=r"(?:leet|1337|leetspeak)$",
        printables_rate=lambda pr: pr)




# -*- coding: UTF-8 -*-
"""ROT Codec - rot-with-N-offset content encoding.

This codec:
- en/decodes strings from str to str
- en/decodes strings from bytes to bytes
- decodes file content to str (read)
- encodes file content from str to bytes (write)
"""
from string import ascii_lowercase as LC, ascii_uppercase as UC

from ..__common__ import *


__examples__ = {
    'enc(rot0|rot--10|rot100)': None,
    'enc(rot1|rot-1|caesar_1)': {'this is a test': "uijt jt b uftu"},
    'enc(rot3|caesar-3)':       {'this is a test': "wklv lv d whvw"},
    'enc(rot47)':               {'this is a test': "E9:D :D 2 E6DE"},
}
__guess__ = ["rot-%d" % i for i in range(1, 26)] + ["rot-47"]


ROT47 = "!\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~"


def _rotn(text, n=13, a=LC+UC, l=26):
    n = n % l
    t = maketrans(a, "".join(a[i:i+l][n:] + a[i:i+l][:n] for i in range(0, len(a), l)))
    return text.translate(t)


def rot_encode(i):
    def encode(text, errors="strict"):
        t = ensure_str(text)
        r = _rotn(t, 47, ROT47, 94) if i == 47 else _rotn(t, i)
        return r, len(r)
    return encode


def rot_decode(i):
    def decode(text, errors="strict"):
        t = ensure_str(text)
        r = _rotn(t, -47, ROT47, 94) if i == 47 else _rotn(t, -i)
        return r, len(r)
    return decode


add("rot", rot_encode, rot_decode, r"(?:caesar|rot)[-_]?([1-9]|1[0-9]|2[0-5]|47)$", aliases=["caesar"],
    printables_rate=lambda pr: pr)




# -*- coding: UTF-8 -*-
"""Codecs extension module.

"""
from __future__ import print_function
from ast import literal_eval
from six import binary_type, text_type

from .__common__ import *
from .__info__ import __author__, __copyright__, __email__, __license__, __source__, __version__


__all__ = ["add", "add_map", "clear", "decode", "encode", "guess", "lookup",  "open", "register", "remove", "reset"]

decode   = codecs.decode
encode   = codecs.encode
guess    = codecs.guess
lookup   = codecs.lookup
open     = codecs.open

_lst = list
list = list_encodings  # not included in __all__ because of shadow name


reset()


def __literal_eval(o):
    """ Non-failing ast.literal_eval alias function. """
    try:
        return literal_eval(str(o))
    except ValueError:
        return literal_eval("'" + str(o) + "'")


def main():
    import argparse, os
    descr = "Codecs Extension (CodExt) {}\n\nAuthor   : {} ({})\nCopyright: {}\nLicense  : {}\nSource   : {}\n" \
            "\nThis tool allows to encode/decode input strings/files with an extended set of codecs.\n\n" \
            .format(__version__, __author__, __email__, __copyright__, __license__, __source__)
    examples = "usage examples:\n- " + "\n- ".join([
        "codext search bitcoin",
        "codext decode base32 -i file.b32",
        "codext encode morse < to_be_encoded.txt",
        "echo \"test\" | codext encode base100",
        "echo -en \"test\" | codext encode braille -o test.braille",
        "codext encode base64 < to_be_encoded.txt > text.b64",
        "echo -en \"test\" | codext encode base64 | codext encode base32",
        "echo -en \"mrdvm6teie6t2cq=\" | codext encode upper | codext decode base32 | codext decode base64",
        "echo -en \"test\" | codext encode upper reverse base32 | codext decode base32 reverse lower",
        "echo -en \"test\" | codext encode upper reverse base32 base64 morse",
        "echo -en \"test\" | codext encode base64 gzip | codext guess",
        "echo -en \"test\" | codext encode base64 gzip | codext guess gzip -c base",
    ])
    parser = argparse.ArgumentParser(description=descr, epilog=examples, formatter_class=argparse.RawTextHelpFormatter)
    sparsers = parser.add_subparsers(dest="command", help="command to be executed")
    parser.add_argument("-i", "--input-file", dest="infile", help="input file (if none, take stdin as input)")
    parser.add_argument("-o", "--output-file", dest="outfile", help="output file (if none, display result to stdout)")
    parser.add_argument("-s", "--strip-newlines", action="store_true", dest="strip",
                        help="strip newlines from input (default: False)")
    encode = sparsers.add_parser("encode", help="encode input using the specified codecs")
    encode.add_argument("encoding", nargs="+", help="list of encodings to apply")
    encode.add_argument("-e", "--errors", default="strict", choices=["ignore", "leave", "replace", "strict"],
                        help="error handling (default: strict)")
    decode = sparsers.add_parser("decode", help="decode input using the specified codecs")
    decode.add_argument("encoding", nargs="+", help="list of encodings to apply")
    decode.add_argument("-e", "--errors", default="strict", choices=["ignore", "leave", "replace", "strict"],
                        help="error handling (default: strict)")
    guess = sparsers.add_parser("guess", help="try guessing the decoding codecs")
    guess.add_argument("encoding", nargs="*", help="list of known encodings to apply (default: none)")
    guess.add_argument("-c", "--codec-categories", nargs="*", help="codec categories to be included in the search ; "
                                                                   "format: string|tuple")
    guess.add_argument("-e", "--exclude-codecs", nargs="*", help="codecs to be explicitely not used ; "
                                                                 "format: string|tuple")
    guess.add_argument("-f", "--stop-function", default="text", help="result checking function (default: text) ; "
                       "format: printables|text|flag|lang_[bigram]|[regex]")
    guess.add_argument("--max-depth", default=5, type=int, help="maximum codec search depth (default: 5)")
    guess.add_argument("--min-depth", default=0, type=int, help="minimum codec search depth before triggering results "
                                                                "(default: 0)")
    guess.add_argument("--extended", action="store_true",
                       help="while using the scoring heuristic, also consider null scores (default: False)")
    guess.add_argument("--heuristic", action="store_true",
                       help="use a scoring heuristic to accelerate guessing (default: False)")
    guess.add_argument("-s", "--do-not-stop", action="store_true",
                       help="do not stop if a valid output is found (default: False)")
    guess.add_argument("-v", "--verbose", action="store_true",
                       help="show guessing information and steps (default: False)")
    rank = sparsers.add_parser("rank", help="rank the most probable encodings based on the given input")
    rank.add_argument("-c", "--codec-categories", help="codec categories to be included in the search ; "
                                                       "format: string|tuple|list(strings|tuples)")
    rank.add_argument("-e", "--exclude-codecs", help="codecs to be explicitely not used ; "
                                                     "format: string|tuple|list(strings|tuples)")
    rank.add_argument("-l", "--limit", type=int, default=10, help="limit the number of displayed results")
    rank.add_argument("--extended", action="store_true",
                      help="while using the scoring heuristic, also consider null scores (default: False)")
    search = sparsers.add_parser("search", help="search for codecs")
    search.add_argument("pattern", nargs="+", help="encoding pattern to search")
    args = parser.parse_args()
    try:
        args.codec_categories = _lst(map(__literal_eval, args.codec_categories))
    except (AttributeError, TypeError):
        pass
    try:
        args.exclude_codecs = _lst(map(__literal_eval, args.exclude_codecs))
    except (AttributeError, TypeError):
        pass
    #print(args.codec_categories, args.exclude_codecs)
    # if a search pattern is given, only handle it
    if args.command == "search":
        results = []
        for enc in args.pattern:
            results.extend(codecs.search(enc))
        print(", ".join(results) or "No encoding found")
        return
    # handle input file or stdin
    c =_input(args.infile)
    c = c.rstrip("\r\n") if isinstance(c, str) else c.rstrip(b"\r\n")
    # strip any other (CR)LF
    if args.strip:
        c = re.sub(r"\r?\n", "", c) if isinstance(c, str) else c.replace(b"\r\n", b"").replace(b"\n", b"")
    if args.command in ["decode", "encode"]:
        # encode or decode
        for encoding in args.encoding:
            c = getattr(codecs, ["encode", "decode"][args.command == "decode"])(c, encoding, args.errors)
        # handle output file or stdout
        if args.outfile:
            with open(args.outfile, 'wb') as f:
                f.write(c)
        else:
            print(ensure_str(c or "Could not decode :-("), end="")
    elif args.command == "guess":
        r = codecs.guess(c,
                         getattr(stopfunc, args.stop_function, args.stop_function),
                         args.min_depth,
                         args.max_depth,
                         args.codec_categories,
                         args.exclude_codecs,
                         args.encoding,
                         not args.do_not_stop,
                         True,  # show
                         args.heuristic,
                         args.extended,
                         args.verbose)
        for i, o in enumerate(r.items()):
            e, out = o
            if len(e) > 0:
                if args.outfile:
                    n, ext = os.path.splitext(args.outfile)
                    fn = args.outfile if len(r) == 1 else "%s-%d%s" % (n, i+1, ext)
                else:
                    print("Codecs: %s" % ", ".join(e))
                    print(ensure_str(out))
        if len(r) == 0:
            print("Could not decode :-(")
    elif args.command == "rank":
        for i, e in codecs.rank(c, args.extended, args.limit, args.codec_categories, args.exclude_codecs):
            s = "[+] %.5f: %s" % (i[0], e)
            print(s if len(s) <= 80 else s[:77] + "...")




# -*- coding: UTF-8 -*-
"""Affine Cipher Codec - affine content encoding.

This codec:
- en/decodes strings from str to str
- en/decodes strings from bytes to bytes
- decodes file content to str (read)
- encodes file content from str to bytes (write)

Reference: https://crypto.interactive-maths.com/affine-cipher.html
"""
from ..__common__ import *


__guess__ = []


def encmap_factory(mask=None):
    mask = mask or "?l?u?s-1,2"
    mask, key = mask.split("-")
    a, b = map(int, key.split(","))
    alphabet = get_alphabet_from_mask(mask)
    encmap = {c: alphabet[(a * alphabet.index(c) + b) % len(alphabet)] for c in alphabet}
    if len(set(encmap.keys())) != len(set(encmap.values())):
        raise LookupError("Bad parameter for encoding 'affine': {}, {}".format(a, b))
    if ' ' not in encmap.keys():
        encmap[' '] = " "
    return encmap


add_map("affine", encmap_factory, pattern=r"affine(?:[-_]cipher)?(?:[-_](.+?\-\d+\,\d+))?$")




# -*- coding: UTF-8 -*-
from argparse import ArgumentParser, RawTextHelpFormatter
from types import MethodType

from .ascii85 import *
from .base45 import *
from .base85 import *
from .base91 import *
from .base100 import *
from .base122 import *
from .baseN import *
from ..__common__ import *
from ..__info__ import __version__


def main():
    descr = """Usage: debase [OPTION]... [FILE]
Base decode multi-layer FILE, or standard input, to standard output.

With no FILE, or when FILE is -, read standard input.

Optional arguments:
  -f, --stop-function   set the result chceking function (default: text)
                         format: printables|text|flag|lang_[bigram]|[regex]
  -i, --ignore-generic  ignore generic base codecs while guess-decoding
  -M, --max-depth       maximum codec search depth (default: 5)
  -m, --min-depth       minimum codec search depth (default: 0)
  -s, --do-not-stop     do not stop if a valid output is found

      --help     display this help and exit
      --verbose  show guessing information and steps
      --version  output version information and exit

Report debase bugs to <https://github.com/dhondta/python-codext/issues/new>
Full documentation at: <https://python-codext.readthedocs.io/en/latest/enc/base.html>
"""
    parser = ArgumentParser(description=descr, formatter_class=RawTextHelpFormatter, add_help=False)
    parser.format_help = MethodType(lambda s: s.description, parser)
    parser.add_argument("file", nargs="?")
    parser.add_argument("-f", "--stop-function", default="text")
    parser.add_argument("-i", "--ignore-generic", action="store_true")
    parser.add_argument("-M", "--max-depth", default=5, type=int)
    parser.add_argument("-m", "--min-depth", default=0, type=int)
    parser.add_argument("-s", "--do-not-stop", action="store_true")
    parser.add_argument("--help", action="help")
    parser.add_argument("--version", action="version")
    parser.add_argument("--verbose", action="store_true")
    parser.version = "CodExt " + __version__
    args = parser.parse_args()
    excl = [[], ["base%d-generic" % i for i in range(2, 255)]][args.ignore_generic]
    sfunc = getattr(stopfunc, args.stop_function, args.stop_function)
    c = _input(args.file)
    c = c.rstrip("\r\n") if isinstance(c, str) else c.rstrip(b"\r\n")
    r = codecs.guess(c, sfunc, args.min_depth, args.max_depth, exclude=excl, codec_categories="base",
                     stop=not args.do_not_stop, show=True, scoring_heuristic=False, debug=args.verbose)
    if not args.do_not_stop:
        print("Could not decode :-(" if len(r) == 0 else ensure_str(list(r.items())[0][1]))




# -*- coding: UTF-8 -*-
"""Ordinal Codec - ordinal content encoding.

This codec:
- en/decodes strings from str to str
- en/decodes strings from bytes to bytes
- decodes file content to str (read)
- encodes file content from str to bytes (write)
"""
from ..__common__ import *


__examples1__ = {
    'enc(ordinal-spaced|ordinals_spaced)': {'this is a test': "116 104 105 115 32 105 115 32 97 32 116 101 115 116"},
}
__examples2__ = {
    'enc(ordinal|ordinals)': {'this is a test': "116104105115032105115032097032116101115116"},
}


ENCMAP1 = {chr(i): str(i) for i in range(256)}
ENCMAP2 = {chr(i): str(i).zfill(3) for i in range(256)}


add_map("ordinal-spaced", ENCMAP1, sep=" ", pattern=r"^ordinals?[-_]spaced$", examples=__examples1__, entropy=3.,
        printables_rate=1.)
add_map("ordinal", ENCMAP2, pattern=r"^ordinals?$", examples=__examples2__, entropy=3., printables_rate=1.)




# -*- coding: UTF-8 -*-
"""Baudot Codec - baudot content conversion to HTML.

This codec:
- en/decodes strings from str to str
- en/decodes strings from bytes to bytes
- decodes file content to str (read)
- encodes file content from str to bytes (write)
"""
from ..__common__ import *


__CODES = ["ccitt1", "ccitt2", "eu", "ita1", "ita2", "ita2_us", "murray", "uk"]
if PY3:
    __CODES.extend(["ita2_meteo", "mtk2"])
__guess__     = ["baudot%s-{}-{}".format(x, y) for x in __CODES for y in ["lsb", "msb"]]
__examples1__ = {
    'enc(baudot-BAD_ALPHABET)': None,
    'enc(baudot_ccitt2_lsb)':   {'TEST 1234': "00001100001010000001001001101111101110011000001010"},
    'enc(baudot-ita1)':         {'TEST 1234': "10101000101010010101100000100000001000100010000101"},
    'enc(baudot_ita2_msb)':     {'TEST 1234': "10000000010010110000001001101110111100110000101010"},
    'enc(baudot-ita2-us)':      {'TEST 1234': "10000000010010110000001001101110111100110000101010"},
    'enc(baudot)':              {'\x01\x02':  None},
    'enc(baudot_ccitt1-lsb)':   {'TEST ':     None},
}
__examples2__ = {
    'enc(baudot_spaced-BAD_ALPHABET)': None,
    'enc(baudot-spaced_ccitt2_lsb)':   {'TEST 1234': "00001 10000 10100 00001 00100 11011 11101 11001 10000 01010"},
    'enc(baudot_spaced-ita1)':         {'TEST 1234': "10101 00010 10100 10101 10000 01000 00001 00010 00100 00101"},
    'enc(baudot-spaced_ita2_msb)':     {'TEST 1234': "10000 00001 00101 10000 00100 11011 10111 10011 00001 01010"},
    'enc(baudot_spaced-ita2-us)':      {'TEST 1234': "10000 00001 00101 10000 00100 11011 10111 10011 00001 01010"},
}
__examples3__ = {
    'enc(baudot_tape-BAD_ALPHABET)': None,
    'enc(baudot_tape-ita1)': {
        'TEST 1234': "***.**\n* *. *\n   .* \n* *.  \n* *. *\n*  .  \n * .  \n   . *\n   .* \n  *.  \n  *. *",
    },
    'dec(baudot-tape_ita2)':       {'BAD_HEADER\n   .* \n': None},
    'dec(baudot-tape_ita2-us)':    {'***.**\nBAD_TAPE\n': None},
    'dec(baudot_tape-ccitt1_lsb)': {'***.**\n   .* \n*  . *\n*  .  \n': None},
}
if PY3:
    __examples1__.update({
        'enc(baudot_ccitt1_lsb)': {'TEST1234':  "101010001010001101010100000100000100000100101"},
        'enc(baudot-fr)':         {'TEST 1234': "10101000101010010101100000100000001000100010000101"},
    })
    __examples2__.update({
        'enc(baudot-spaced_ccitt1_lsb)': {'TEST1234':  "10101 00010 10001 10101 01000 00100 00010 00001 00101"},
        'enc(baudot_spaced-fr)':         {'TEST 1234': "10101 00010 10100 10101 10000 01000 00001 00010 00100 00101"},
    })


PATTERN = r"^baudot%s([-_](?:ccitt1|ccitt2|eu|fr|ita1|ita2|ita2[-_](?:us" + (r"|meteo" if PY3 else r"") + r")" + \
          (r"|mtk2" if PY3 else r"") + r"|murray|uk|us_tty)(?:[-_](?:lsb|msb))?)?$"
# reserved character
RES_CHR = "\xff"

# sources:
# - http://rabbit.eng.miami.edu/info/baudot.html
# - https://en.wikipedia.org/wiki/Baudot_code
# - https://fr.qwe.wiki/wiki/Baudot_code
# all alphabets consider MSB by default
# CCITT-1 original Baudot code (source: http://rabbit.eng.miami.edu/info/baudot.html)
CCITT1 = [
    "00001", "00010",
    "\x00\xff\xff\xffA-JKEXGM/ZHLYSBRUTCQIWFNOVDP",
    "\x00\xff\xff\xff1.6(2\xff7)\xff:\xff=3\xff8-4\xff9/\xff?\xff£5'0+" if PY3 else \
        "\x00\xff\xff\xff1.6(2\xff7)\xff:\xff=3\xff8-4\xff9/\xff?\xff$5'0+",
]
# CCITT-2 revised Baudot code (source: http://rabbit.eng.miami.edu/info/baudot.html)
CCITT2 = [
    "11111", "11011",
    "\x00E\nA SIU\rDRJNFCKTZLWHYPQOBG\xffMXV\xff",
    "\x003\n- \x0787\r$4',!:(5\")2#6019?&\xff./;\xff",
]
# Original Baudot (French/European ; sources: https://fr.qwe.wiki/wiki/Baudot_code
#                                             https://en.wikipedia.org/wiki/Baudot_code)
BAUDOT = EU = FR = [
    "10000", "01000",
    "\x00AEÉYUIO\xffJGHBCFD \nXZSTWV\x7fKMLRQNP" if PY3 else "\x00AEeYUIO\xffJGHBCFD \nXZSTWV\x7fKMLRQNP",
    "\x0012&34°5 67h89f0\xff.,:;!?'\x7f()=-/\u2116%" if PY3 else "\x0012&34o5 67h89f0\xff.,:;!?'\x7f()=-/\xff%",
]
# International Telegraphic Alphabet 1 (sources: https://fr.qwe.wiki/wiki/Baudot_code
#                                                https://en.wikipedia.org/wiki/Baudot_code)
ITA1 = [
    "10000", "01000",
    "\x00AE\rYUIO\xffJGHBCFD \xffXZSTWV\x7fKMLRQNP",
    "\x0012\r34\xff5 67+89\xff0\xff\n,:.\xff?'\x7f()=-/\xff%",
]
# International Telegraphic Alphabet 2 (sources: https://fr.qwe.wiki/wiki/Baudot_code
#                                                https://en.wikipedia.org/wiki/Baudot_code)
ITA2 = [
    "11111", "11011",
    "\x00E\nA SIU\rDRJNFCKTZLWHYPQOBG\xffMXV\xff",
    "\x003\n- '87\r\x054\x07,!:(5+)2$6019?&\xff./=\xff",
]
# International Telegraphic Alphabet 2 - US TTY (sources: https://fr.qwe.wiki/wiki/Baudot_code
#                                                         https://en.wikipedia.org/wiki/Baudot_code)
ITA2_US = US_TTY = [
    "11111", "11011",
    "\x00E\nA SIU\rDRJNFCKTZLWHYPQOBG\xffMXV\xff",
    "\x003\n- \x0787\r$4',!:(5\")2#6019?&\xff./;\xff",
]
# International Telegraphic Alphabet 2 - Meteo version (source: https://en.wikipedia.org/wiki/Baudot_code)
if PY3:
    ITA2_METEO = [
        "11111", "11011",
        "\x00E\nA SIU\rDRJNFCKTZLWHYPQOBG\xffMXV\xff",
        "-3\n\u2191 \x0787\r\u21974\u2199\u29b7\u2192\u25ef\u21905+\u21962\u21936019\u2295\u2198\xff./\u29b6\xff",
    ]
# Russian MTK-2 alphabet (source: https://fr.qwe.wiki/wiki/Baudot_code)
if PY3:
    MTK2 = [
        "11111", "11011",
        "\x00Е\n\xff СИУ\r\xffРЙНФЦКТЗЛВХЫПЯОБГ\xffМЬЖ\xff",
        "\x003\n- '87\r\xff4Ю,Э:(5+)2Щ6019?Ш\xff./=\xff",
    ]
# Murray code ; NB: not all fractions are supported (source: https://en.wikipedia.org/wiki/Baudot_code)
MURRAY = [
    "00100", "11011",
    " E\xffA\xffSIU\nDRJNFCKTZLWHYPQOBF\xffMXV\x7f", 
    "\x003\xff\xff\xff'87\n²4\xff-\u215f(\xff5./2\xff6019?\xff\xff,£)*" if PY3 else \
        "\x003\xff\xff\xff'87\n²4\xff-\u215f(\xff5./2\xff6019?\xff\xff,$)*",
]
# English Baudot ; NB: not all fractions are supported (sources: https://fr.qwe.wiki/wiki/Baudot_code
#                                                                https://en.wikipedia.org/wiki/Baudot_code)
UK = [
    "10000", "01000",
    "\x00AE/YUIO\xffJGHBCFD -XZSTWV\x7fKMLRQNP", 
    "\x0012\u215f34\xff5 67\xb989\xff0\xff.\xff:\xff²?'\x7f()=-/£+" if PY3 else \
        "\x0012\xff34\xff5 67\xb989\xff0\xff.\xff:\xff²?'\x7f()=-/$+",
]


def _bits_from_tape(tape, trans={'*': "1", ' ': "0"}):
    """ Converts a tape-like string with the given translation for ones and zeros to a series of bits. """
    bits = ""
    trans_rev = {v: k for k, v in trans.items()}
    for i, line in enumerate(tape.splitlines()):
        if i == 0:
            if line != trans_rev['1'] * 3 + "." + trans_rev['1'] * 2:
                raise ValueError("Bad tape header '{}'".format(line))
        else:
            line = line[:3] + line[4:]
            if len(line) != 5:
                raise ValueError("Bad tape line '{}'".format(line))
            bits += "".join(trans.get(c, "") for c in line)
    return bits


def _bits_to_tape(bits, trans={'1': "*", '0': " "}):
    """ Converts a series of bits to a tape-like string with the given translation for ones and zeros. """
    tape = [trans['1'] * 3 + "." + trans['1'] * 2]
    for i in range(0, len(bits), 5):
        group = "".join(trans[b] for b in bits[i:i+5])
        tape.append(group[:3] + "." + group[3:])
    return "\n".join(tape)


def _check_alphabet(alphabet):
    """ Checks the length of letters and figures (must be 32 chars). """
    for chars in alphabet:
        l = len(chars)
        if l != 32:
            raise ValueError("Bad length of alphabet (%d instead of 32)" % l)


def _handle_alphabet(alphabet):
    """ Gets the given alphabet name and transforms it to its dictionary with letters and figures. """
    alphabet = (alphabet or "baudot").lower().replace("-", "_").strip("_")
    if "_lsb" in alphabet:
        alphabet = alphabet.replace("_lsb", "")
        func = lambda x: x[::-1]
    else:
        alphabet = alphabet.replace("_msb", "")
        func = lambda x: x
    _ = globals()[alphabet.upper()]
    st, a = _[:2], _[2:]
    _check_alphabet(a)
    alphabet = {n: {ch: bin(i)[2:].zfill(5) for i, ch in enumerate(alph) if ch != RES_CHR} for n, alph in \
                zip(["letters", "figures"], a)}
    return alphabet, {'letters': st[0], 'figures': st[1]}, func


def baudot_encode(alphabet=None, spaced=False, tape=False):
    ename = "baudot" + ("-spaced" if spaced else "-tape" if tape else "")
    alphabet, states, func = _handle_alphabet(alphabet)
    def encode(text, errors="strict"):
        text = text.upper()
        s, l, state, seen_states = "", len(b(text)), None, []
        for i, c in enumerate(text):
            # if the state is undefined yet, find the relevant alphabet
            if state is None:
                bits= None
                for st in states.keys():
                    try:
                        bits = func(alphabet[st][c])
                        state = st
                        if st not in seen_states:
                            seen_states.append(st)
                        break
                    except KeyError:
                        pass
                if bits is None:
                    bits = handle_error(ename, errors, "?", 5)(c, i)
                s += bits
            # otherwise, handle state change (when the current alphabet does not contain the character to encode but the
            #  other alphabet does
            else:
                try:
                    s += func(alphabet[state][c])
                    continue
                except KeyError:
                    state = list(set(states.keys()) - {state})[0]
                try:
                    s += func(states[state]) + func(alphabet[state][c])
                    if state not in seen_states:
                        seen_states.append(state)
                except KeyError as e:
                    state = list(set(states.keys()) - {state})[0]  # reset the state
                    s += handle_error(ename, errors, "?", 5)(c, i)
        # by default, if no state is specified, the encoded string is handled as letters ; so if figures are used only,
        #  it is necessary to include the groups of bits for figures at the beginning of the encoded string
        s = (states['figures'] if seen_states == ["figures"] else "") + s
        if spaced:
            s = " ".join(s[i:i+5] for i in range(0, len(s), 5))
        elif tape:
            s = _bits_to_tape(s)
        return s, l
    return encode


def baudot_decode(alphabet=None, spaced=False, tape=False):
    ename = "baudot" + ("-spaced" if spaced else "-tape" if tape else "")
    alphabet, states, func = _handle_alphabet(alphabet)
    alphabet = {st: {v: k for k, v in alph.items()} for st, alph in alphabet.items()}
    states = {v: k for k, v in states.items()}
    def decode(text, errors="strict"):
        s, l = "", len(b(text))
        if spaced:
            text = text.replace(" ", "")
        elif tape:
            text = _bits_from_tape(text)
        # infer the starting state by searching for the first encountered groups of bits indicating a valid state ;
        #  by default, we assume letters
        state = "letters"
        for i in range(0, len(text), 5):
            bits = func(text[i:i+5])
            # the following code handles a possible ambiguity ; e.g. when letters have a group of bits matching
            #  a state change
            if bits in states.keys():
                error = False
                # so, when we see the bits of a state, we parse previous groups in order to determine if they are valid
                #  groups in the corresponding state, that is, if no error occurs ; if an error occurs, then it is a
                #  valid state change and not simply a character, and we can set it as the starting state
                for j in range(i-5, 0, -5):
                    try:
                        alphabet[states[bits]][text[j:j+5]]
                    except KeyError:
                        error = True
                        break
                if error:
                    state = list(set(states.values()) - {states[bits]})[0]
                    break
        # now parse the input text
        for i in range(0, len(text), 5):
            bits = func(text[i:i+5])
            try:
                s += alphabet[state][bits]
            except KeyError:
                if bits in states.keys() and states[bits] != state:
                    state = states[bits]
                else:
                    s += handle_error(ename, errors, decode=True, item="group")(bits, i//5)
        return s, l
    return decode


add("baudot", baudot_encode, baudot_decode, PATTERN % r"", examples=__examples1__, guess=[x % "" for x in __guess__],
    entropy=1., printables_rate=1.)


baudot_spaced_encode = lambda a: baudot_encode(a, spaced=True)
baudot_spaced_decode = lambda a: baudot_decode(a, spaced=True)
add("baudot-spaced", baudot_spaced_encode, baudot_spaced_decode, PATTERN % r"[-_]spaced", examples=__examples2__,
    guess=[x % "-spaced" for x in __guess__], entropy=1.48, printables_rate=1.)


baudot_tape_encode = lambda a: baudot_encode(a, tape=True)
baudot_tape_decode = lambda a: baudot_decode(a, tape=True)
add("baudot-tape", baudot_tape_encode, baudot_tape_decode, PATTERN % r"[-_]tape", examples=__examples3__,
    guess=[x % "-tape" for x in __guess__], entropy=1.86, printables_rate=1.)




# -*- coding: UTF-8 -*-
"""Klopf Codec - Polybius-based content encoding.

This codec:
- en/decodes strings from str to str
- en/decodes strings from bytes to bytes
- decodes file content to str (read)
- encodes file content from str to bytes (write)
"""
from ..__common__ import *


__examples__ = {
    'enc(klopf|klopfcode)': {'this is a test': "44324234 4234 11 44513444"},
}


ENCMAP = {"ABCDEFGHIKLMNOPQRSTUVWXYZ"[y*5+x]: "".join([str(x+1), str(y+1)]) for x in range(5) for y in range(5)}
ENCMAP['J'] = "43"
ENCMAP[' '] = " "


add_map("klopf", ENCMAP, ignore_case="both", pattern=r"^(?:klopf(?:code)?)$", printables_rate=1.)




# -*- coding: UTF-8 -*-
"""Atbash Cipher Codec - atbash content encoding.

This codec:
- en/decodes strings from str to str
- en/decodes strings from bytes to bytes
- decodes file content to str (read)
- encodes file content from str to bytes (write)

Reference: https://crypto.interactive-maths.com/atbash-cipher.html
"""
from ..__common__ import *


__guess__ = ["atbash"]


def encmap_factory(mask=None):
    mask = mask or "?u?l"
    # [...] enclosure causes the mask to be handled as a whole
    if mask[0] == "[" and mask[-1] == "]":
        alphabet = get_alphabet_from_mask(mask[1:-1])
        return {k: v for k, v in zip(alphabet, alphabet[::-1])}
    # not enclosing the whole mask means that each group is to be considered separately
    else:
        m = {}
        for group in re.findall(r"(\?.|[^?]+)", mask):
            alphabet = get_alphabet_from_mask(group)
            m.update({k: v for k, v in zip(alphabet, alphabet[::-1])})
        return m


add_map("atbash", encmap_factory, no_error=True, pattern=r"atbash(?:[-_]cipher)?(?:[-_](.+))?$")

